"""Classical socket implementation using a hub that is shared across threads.

This module contains the ThreadSocket class which is an implementation of the Socket
interface that can be used by Hosts that run in a multi-thread simulation.
"""

from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from netqasm.logging.glob import get_netqasm_logger
from netqasm.logging.output import ClassCommLogger, SocketOperation
from netqasm.sdk.classical_communication.message import StructuredMessage
from netqasm.sdk.config import LogConfig
from netqasm.util.log import LineTracker

from ..socket import Socket
from .socket_hub import _socket_hub, _SocketHub

if TYPE_CHECKING:
    import logging

T_ThreadSocketKey = Tuple[str, str, int]


def trim_msg(msg: str) -> str:
    trimmed_msg = msg
    if trimmed_msg.endswith("EOF"):
        trimmed_msg = trimmed_msg.split("EOF")[0]
    return trimmed_msg


def log_send(method):
    def new_method(self, msg: str) -> None:
        hln = None
        hfl = None
        if self._line_tracker is not None:
            hostline = self._line_tracker.get_line()
            if hostline is not None:
                hln = hostline.lineno
                hfl = hostline.filename

        if self._comm_logger is not None:
            trimmed_msg = trim_msg(msg)
            log = f"Send classical message to {self.remote_app_name}: {trimmed_msg}"
            self._comm_logger.log(
                socket_op=SocketOperation.SEND,
                msg=trimmed_msg,
                sender=self._app_name,
                receiver=self._remote_app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        method(self, msg)

    return new_method


def log_send_structured(method):
    def new_method(self, msg: StructuredMessage):
        hln = None
        hfl = None
        if self._line_tracker is not None:
            hostline = self._line_tracker.get_line()
            if hostline is not None:
                hln = hostline.lineno
                hfl = hostline.filename

        logged_msg = f"{msg.header}: {msg.payload}"
        raw_msg = json.dumps(msg.__dict__)

        if self._comm_logger is not None:
            log = f"Send classical message to {self.remote_app_name}: {raw_msg}"
            self._comm_logger.log(
                socket_op=SocketOperation.SEND,
                msg=logged_msg,
                sender=self._app_name,
                receiver=self._remote_app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        method(self, raw_msg)

    return new_method


def log_recv(method):
    def new_method(self, *args, **kwargs):
        hln = None
        hfl = None
        if self._line_tracker is not None:
            hostline = self._line_tracker.get_line()
            if hostline is not None:
                hln = hostline.lineno
                hfl = hostline.filename

        if self._comm_logger is not None:
            log = f"Waiting for a classical message from {self.remote_app_name}..."
            self._comm_logger.log(
                socket_op=SocketOperation.WAIT_RECV,
                msg=None,
                sender=self._remote_app_name,
                receiver=self._app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        msg = method(self, *args, **kwargs)

        if self._comm_logger is not None:
            trimmed_msg = trim_msg(msg)
            log = f"Message received from {self.remote_app_name}: {trimmed_msg}"
            self._comm_logger.log(
                socket_op=SocketOperation.RECV,
                msg=trimmed_msg,
                sender=self._remote_app_name,
                receiver=self._app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        return msg

    return new_method


def log_recv_structured(method):
    def new_method(self, *args, **kwargs):
        hln = None
        hfl = None
        if self._line_tracker is not None:
            hostline = self._line_tracker.get_line()
            if hostline is not None:
                hln = hostline.lineno
                hfl = hostline.filename

        if self._comm_logger is not None:
            log = f"Waiting for a classical message from {self.remote_app_name}..."
            self._comm_logger.log(
                socket_op=SocketOperation.WAIT_RECV,
                msg=None,
                sender=self._remote_app_name,
                receiver=self._app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        raw_msg = method(self, *args, **kwargs)
        msg_dict = json.loads(raw_msg)
        msg = StructuredMessage(header=msg_dict["header"], payload=msg_dict["payload"])
        logged_msg = f"{msg.header}: {msg.payload}"

        if self._comm_logger is not None:
            log = f"Message received from {self.remote_app_name}: {msg}"
            self._comm_logger.log(
                socket_op=SocketOperation.RECV,
                msg=logged_msg,
                sender=self._remote_app_name,
                receiver=self._app_name,
                socket_id=self._id,
                hln=hln,
                hfl=hfl,
                log=log,
            )

        return msg

    return new_method


class ThreadSocket(Socket):
    """Classical socket implementation for multi-threaded simulations.

    This implementation should be used when the simulation consists of a single process,
    with a thread for each Host. A global SocketHub instance must be available and
    shared between all threads. The ThreadSocket instance for a Host can then 'send'
    and 'receive' messages to/from other Hosts by writing/reading from the shared
    SocketHub memory.

    Currently only used with the SquidASM simulator backend.
    """

    _COMM_LOGGERS: Dict[str, Optional[ClassCommLogger]] = {}
    _SOCKET_HUB: _SocketHub = _socket_hub

    def __init__(
        self,
        app_name: str,
        remote_app_name: str,
        socket_id: int = 0,
        timeout: Optional[float] = None,
        use_callbacks: bool = False,
        log_config: Optional[LogConfig] = None,
    ):
        """ThreadSocket constructor.

        :param app_name: application/Host name of this socket's owner
        :param remote_app_name: remote application/Host name
        :param socket_id: local ID to use for this socket
        :param timeout: maximum amount of real time to try to connect to the remote
            socket
        :param use_callbacks: whether to call the `recv_callback` method upon receiving
            a message
        :param log_config: logging configuration for this socket
        """
        self._app_name: str = app_name
        self._remote_app_name: str = remote_app_name
        self._id: int = socket_id
        if app_name == remote_app_name:
            raise ValueError(
                f"Cannot connect to itself app_name {app_name} = remote_app_name {remote_app_name}"
            )

        if log_config is None:
            log_config = LogConfig()

        self._line_tracker: LineTracker = LineTracker(log_config=log_config)
        self._track_lines: bool = log_config.track_lines

        # Use callbacks
        self._use_callbacks: bool = use_callbacks

        # Received messages
        # TODO: remove?
        self._received_messages: List[str] = []

        # Logger
        self._logger: logging.Logger = get_netqasm_logger(
            f"{self.__class__.__name__}{self.key}"
        )

        self._logger.debug("Setting up connection")

        # Classical communication logger
        self._comm_logger: Optional[ClassCommLogger]
        if log_config.comm_log_dir is None:
            self._comm_logger = None
        else:
            self._comm_logger = self.__class__.get_comm_logger(
                app_name=self.app_name,
                comm_log_dir=log_config.comm_log_dir,
            )

        # Connect
        self._SOCKET_HUB.connect(self, timeout=timeout)

    @classmethod
    def get_comm_logger(cls, app_name: str, comm_log_dir: str) -> ClassCommLogger:
        comm_logger = cls._COMM_LOGGERS.get(app_name)
        if comm_logger is None:
            filename = f"{str(app_name).lower()}_class_comm.yaml"
            filepath = os.path.join(comm_log_dir, filename)
            comm_logger = ClassCommLogger(filepath=filepath)
            cls._COMM_LOGGERS[app_name] = comm_logger
        return comm_logger

    def __del__(self) -> None:
        if self.connected:
            self._logger.debug("Closing connection")
        self._connected = False
        self._SOCKET_HUB.disconnect(self)

    @property
    def app_name(self) -> str:
        return self._app_name

    @property
    def remote_app_name(self) -> str:
        return self._remote_app_name

    @property
    def id(self) -> int:
        return self._id

    @property
    def key(self) -> T_ThreadSocketKey:
        return self.app_name, self.remote_app_name, self.id

    @property
    def remote_key(self) -> T_ThreadSocketKey:
        return self.remote_app_name, self.app_name, self.id

    @property
    def connected(self) -> bool:
        return self._SOCKET_HUB.is_connected(self)

    @property
    def use_callbacks(self) -> bool:
        return self._use_callbacks

    @use_callbacks.setter
    def use_callbacks(self, value: bool) -> None:
        self._use_callbacks = value

    @log_send
    def send(self, msg: str) -> None:
        """Sends a message to the remote node.

        Parameters
        ----------
        msg : str
            Message to be sent.

        Raises
        ------
        ConnectionError
            If the remote connection is unresponsive.
        """
        if not isinstance(msg, str):
            raise TypeError(f"Messages needs to be a string, not {type(msg)}")
        if not self.connected:
            raise ConnectionError("Socket is not connected so cannot send")

        self._SOCKET_HUB.send(self, msg)

    @log_recv
    def recv(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
        maxsize: Optional[int] = None,
    ) -> str:
        """Receive a message form the remote node.

        If block is True the method will block until there is a message or a timeout is reached.
        Otherwise the method will raise a `RuntimeError` if there is not message to receive directly.

        Parameters
        ----------
        block : bool
            Whether to block for an available message
        timeout : float, optional
            Optionally use a timeout for trying to recv a message. Only used if `block=True`.
        maxsize : int
            How many bytes to maximally receive (not used here)

        Returns
        -------
        str
            The message received

        Raises
        ------
        RuntimeError
            If `block=False` and there is no available message
        """
        # TODO use maxsize?
        msg = self._SOCKET_HUB.recv(self, block=block, timeout=timeout)
        if not isinstance(msg, str):
            raise RuntimeError(f"Received message of type {type(msg)} instead of str")
        return msg

    @log_send_structured
    def send_structured(self, msg: StructuredMessage) -> None:
        if not self.connected:
            raise ConnectionError("Socket is not connected so cannot send")

        self._SOCKET_HUB.send(self, msg)

    @log_recv_structured
    def recv_structured(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
        maxsize: Optional[int] = None,
    ) -> StructuredMessage:
        # TODO use maxsize?
        msg = self._SOCKET_HUB.recv(self, block=block, timeout=timeout)
        # if not isinstance(msg, StructuredMessage):
        #     raise RuntimeError(f"Received message of type {type(msg)} instead of StructuredMessage")
        # TODO fix return value type hints
        return msg  # type: ignore

    def wait(self) -> None:
        """Waits until the connection gets lost"""
        while True:
            if not self.connected:
                return

    def send_silent(self, msg: str) -> None:
        """Sends a message without logging"""
        if not isinstance(msg, str):
            raise TypeError(f"Messages needs to be a string, not {type(msg)}")
        if not self.connected:
            raise ConnectionError("Socket is not connected so cannot send")

        self._SOCKET_HUB.send(self, msg)

    def recv_silent(
        self, block: bool = True, timeout: Optional[float] = None, maxsize: int = None
    ) -> str:
        """Receive a message without logging"""
        msg = self._SOCKET_HUB.recv(self, block=block, timeout=timeout)
        if not isinstance(msg, str):
            raise RuntimeError(f"Received message of type {type(msg)} instead of str")
        return msg


class StorageThreadSocket(ThreadSocket):
    def __init__(self, app_name: str, remote_app_name: str, **kwargs):
        """ThreadSocket that simply stores any message comming in"""
        self._storage: List[str] = []
        super().__init__(app_name, remote_app_name, use_callbacks=True, **kwargs)

    def recv_callback(self, msg: str) -> None:
        self._storage.append(msg)



from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection
from netqasm.sdk.toolbox.measurements import parity_meas

logger = get_netqasm_logger()


def _get_default_strategy():
    return [
        ["XI", "-XZ", "IZ"],  # col 0
        ["XX", "YY", "ZZ"],  # col 1
        ["IX", "-ZX", "ZI"],  # col 2
    ]


def main(app_config=None, col=0, strategy=None):

    # Get the strategy
    if strategy is None:
        strategy = _get_default_strategy()
    if col >= len(strategy):
        raise ValueError(f"Not a col in the square {col}")

    # Create a EPR socket for entanglement generation
    epr_socket = EPRSocket("player1")

    # Initialize the connection
    player2 = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )
    with player2:
        # Create EPR pairs
        q1 = epr_socket.recv_keep()[0]
        q2 = epr_socket.recv_keep()[0]

        player2.flush()

        # Make sure we order the qubits consistently with Player1
        # Get entanglement IDs
        q1_ID = q1.entanglement_info.sequence_number
        q2_ID = q2.entanglement_info.sequence_number

        if int(q1_ID) < int(q2_ID):
            qb = q1
            qd = q2
        else:
            qb = q2
            qd = q1

        # Perform the three measurements
        m0, m1, m2 = (parity_meas([qb, qd], strategy[col][i]) for i in range(3))

    to_print = "\n\n"
    to_print += "==========================\n"
    to_print += "App player2: column is:\n"
    to_print += "(" + "_" * col + str(m0) + "_" * (2 - col) + ")\n"
    to_print += "(" + "_" * col + str(m1) + "_" * (2 - col) + ")\n"
    to_print += "(" + "_" * col + str(m2) + "_" * (2 - col) + ")\n"
    to_print += "==========================\n"
    to_print += "\n\n"
    logger.info(to_print)

    return {
        "col": [int(m0), int(m1), int(m2)],
    }


if __name__ == "__main__":
    main()



import math

from netqasm.logging.output import get_new_app_logger
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state
from netqasm.sdk.toolbox import set_qubit_state
from netqasm.sdk.toolbox.sim_states import qubit_from, to_dm
from netqasm.util.states import bloch_sphere_rep


def main(app_config=None, phi=0.0, theta=0.0):
    phi *= math.pi
    theta *= math.pi

    log_config = app_config.log_config

    # socket for creating an EPR pair with Controller
    controller_epr = EPRSocket("controller")

    # socket for communicating classical messages with Controller
    class_socket = Socket("target", "controller", log_config=app_config.log_config)

    app_logger = get_new_app_logger(app_name="sender", log_config=log_config)

    # connect to the back-end
    target = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[controller_epr],
    )

    with target:
        app_logger.log("Creating EPR pair with controller...")
        # create one EPR pair with Controller
        epr = controller_epr.recv(1)[0]

        # initialize target qubit of the distributed CNOT
        target.flush()
        app_logger.log("Initializing target qubit...")
        target_qubit = Qubit(target)
        set_qubit_state(target_qubit, phi, theta)

        # let back-end execute the quantum operations above
        target.flush()
        app_logger.log("Initialized target qubit")

        class_socket.send_silent("sync")

        # wait for Controller's measurement outcome
        m = class_socket.recv()

        # if outcome = 1, apply an X gate on the local EPR half
        if m == "1":
            app_logger.log("Outcome = 1, so doing X correction")
            epr.X()
        else:
            app_logger.log("Outcome = 0, no correction needed")

        # At this point, `epr` is correlated with the control qubit on Controller's side.
        # (If Controller's control was in a superposition, `epr` is now entangled with it.)
        # Use `epr` as the control of a local CNOT on the target qubit.
        epr.cnot(target_qubit)

        target.flush()

        # undo any potential entanglement between `epr` and Controller's control qubit
        app_logger.log("Undo entanglement between control and EPR qubit")
        epr.H()
        epr_meas = epr.measure()
        target.flush()

        # Controller will do a controlled-Z based on the outcome to undo the entanglement
        class_socket.send(str(epr_meas))

        # Wait for an ack before exiting
        assert class_socket.recv_silent() == "ACK"

        if get_simulator() == Simulator.NETSQUID:
            original_dm = to_dm(qubit_from(phi, theta))
            final_dm = get_qubit_state(target_qubit, reduced_dm=True)
            print(bloch_sphere_rep(final_dm))
            return {
                "epr_meas": int(epr_meas),
                "original_state": original_dm.tolist(),
                "final_state": final_dm if final_dm is None else final_dm.tolist(),
            }
        else:
            return {"epr_meas": int(epr_meas)}



"""Qubit representation.

This module contains the `Qubit` class, which are used by application scripts
as handles to in-memory qubits.
"""
from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, Optional, Tuple, Union

from netqasm.lang.ir import GenericInstr
from netqasm.sdk.futures import Future, RegFuture

if TYPE_CHECKING:
    from netqasm import qlink_compat
    from netqasm.lang.operand import Template
    from netqasm.sdk import connection as sdkconn
    from netqasm.sdk.builder import Builder


class QubitNotActiveError(MemoryError):
    pass


class QubitMeasureBasis(Enum):
    X = 0
    Y = auto()
    Z = auto()


class Qubit:
    """Representation of a qubit that has been allocated in the quantum node.

    A `Qubit` instance represents a quantum state that is stored in a physical qubit
    somewhere in the quantum node.
    The particular qubit is identified by its virtual qubit ID.
    To which physical qubit ID this is mapped (at a given time), is handled completely
    by the quantum node controller and is not known to the `Qubit` itself.

    A `Qubit` object can be instantiated in an application script.
    Such an instantiation is automatically compiled into NetQASM instructions that
    allocate and initialize a new qubit in the quantum node controller.

    A `Qubit` object may also be obtained by SDK functions that return them, like
    the `create()` method on an `EPRSocket`, which returns the object as a handle to
    the qubit that is now entangled with one in another node.

    Qubit operations like applying gates and measuring them are done by calling
    methods on a `Qubit` instance.
    """

    def __init__(
        self,
        conn: sdkconn.BaseNetQASMConnection,
        add_new_command: bool = True,
        ent_info: Optional[qlink_compat.LinkLayerOKTypeK] = None,
        virtual_address: Optional[int] = None,
    ):
        """Qubit constructor. This is the standard way to allocate a new qubit in
        an application.

        :param conn: connection of the application in which to allocate the qubit
        :param add_new_command: whether to automatically add NetQASM instructions to
            the current subroutine to allocate and initialize the qubit
        :param ent_info: entanglement generation information in case this qubit is
            the result of an entanglement generation request
        :param virtual_address: explicit virtual ID to use for this qubit. If None,
            a free ID is automatically chosen.
        """
        self._conn: sdkconn.BaseNetQASMConnection = conn
        if virtual_address is None:
            self._qubit_id: int = self.builder.new_qubit_id()
        else:
            self._qubit_id = virtual_address

        if add_new_command:
            self.builder._build_cmds_new_qubit(qubit_id=self.qubit_id)

        self._active: bool = False
        self._activate()

        self._ent_info: Optional[qlink_compat.LinkLayerOKTypeK] = ent_info

        self._remote_ent_node: Optional[str] = None

    def __str__(self) -> str:
        if self.active:
            return "Qubit at the node {}".format(self._conn.node_name)
        else:
            return "Not active qubit"

    @property
    def connection(self) -> sdkconn.BaseNetQASMConnection:
        """Get the NetQASM connection of this qubit"""
        return self._conn

    @property
    def builder(self) -> Builder:
        """Get the Builder of this qubit's connection"""
        return self.connection.builder

    @property
    def qubit_id(self) -> int:
        """Get the qubit ID"""
        return self._qubit_id

    @qubit_id.setter
    def qubit_id(self, qubit_id: int) -> None:
        assert isinstance(qubit_id, int), "qubit_id should be an int"
        self._qubit_id = qubit_id

    @property
    def active(self) -> bool:
        return self._active

    @active.setter
    def active(self, active: bool) -> None:
        assert isinstance(active, bool), "active shoud be a bool"

        # Check if not already new state
        if self._active == active:
            return

        self._active = active

        if active:
            self._activate()
        else:
            self._deactivate()

    def _activate(self) -> None:
        self._active = True
        if not self.builder._mem_mgr.is_qubit_active(self):
            self.builder._mem_mgr.activate_qubit(self)

    def _deactivate(self) -> None:
        self._active = False
        if self.builder._mem_mgr.is_qubit_active(self):
            self.builder._mem_mgr.deactivate_qubit(self)

    @property
    def entanglement_info(self) -> Optional[qlink_compat.LinkLayerOKTypeK]:
        """Get information about the successful link layer request that resulted in
        this qubit."""
        return self._ent_info

    @property
    def remote_entangled_node(self) -> Optional[str]:
        """Get the name of the remote node the qubit is entangled with.

        If not entanled, `None` is returned.
        """
        if self._remote_ent_node is not None:
            return self._remote_ent_node
        if self.entanglement_info is None:
            return None
        # Lookup remote entangled node
        remote_node_id = self.entanglement_info.remote_node_id
        remote_node_name = self._conn.network_info._get_node_name(
            node_id=remote_node_id
        )
        self._remote_ent_node = remote_node_name
        return remote_node_name

    def assert_active(self) -> None:
        """Assert that the qubit is active, i.e. allocated."""
        if not self.active:
            raise QubitNotActiveError(f"Qubit {self.qubit_id} is not active")

    def measure(
        self,
        future: Optional[Union[Future, RegFuture]] = None,
        inplace: bool = False,
        store_array: bool = True,
        basis: QubitMeasureBasis = QubitMeasureBasis.Z,
        basis_rotations: Optional[Tuple[int, int, int]] = None,
    ) -> Union[Future, RegFuture]:
        """
        Measure the qubit in the standard basis and get the measurement outcome.

        :param future: the `Future` to place the outcome in. If None, a Future
            is created automatically.
        :param inplace: If False, the measurement is destructive and the qubit
            is removed from memory. If True, the qubit is left in the
            post-measurement state.
        :param store_array: whether to store the outcome in an array. If not, it
            is placed in a register. Only used if `future` is None.
        :param basis: in which of the Pauli bases (X, Y or Z) to measure.
            Default is Z. Ignored if `basis_rotations` is not None.
        :param basis_rotations: rotations to apply before measuring in the
            Z-basis. This can be used to specify arbitrary measurement bases.
            The 3 values are interpreted as 3 rotation angles, for an X- Y-, and
            another X-rotation, respectively. Each angle is interpreted as a
            multiple of pi/16. For example, if `basis_rotations` is (8, 0, 0),
            an X-rotation is applied with angle 8*pi/16 = pi/2 radians, followed
            by a Y-rotation of angle 0 and an X-rotation of angle 0. Finally,
            the measurement is done in the Z-basis.
        :return: the Future representing the measurement outcome. It is a
            `Future` if
        the result is in an array (default) or `RegFuture` if the result is in a
        register.
        """
        self.assert_active()

        if future is None:
            if store_array:
                array = self.builder.alloc_array(1)
                future = array.get_future_index(0)
            else:
                future = RegFuture(self._conn)

        self.builder._build_cmds_measure(
            qubit_id=self.qubit_id,
            future=future,
            inplace=inplace,
            basis=basis,
            rotations=basis_rotations,
        )

        if not inplace:
            self.active = False

        return future

    def X(self) -> None:
        """Apply an X gate on the qubit."""
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.X, qubit_id=self.qubit_id
        )

    def Y(self) -> None:
        """Apply a Y gate on the qubit."""
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.Y, qubit_id=self.qubit_id
        )

    def Z(self) -> None:
        """Apply a Z gate on the qubit."""
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.Z, qubit_id=self.qubit_id
        )

    def T(self) -> None:
        """Apply a T gate on the qubit.

        A T gate is a Z-rotation with angle pi/4.
        """
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.T, qubit_id=self.qubit_id
        )

    def H(self) -> None:
        """Apply a Hadamard gate on the qubit."""
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.H, qubit_id=self.qubit_id
        )

    def K(self) -> None:
        """Apply a K gate on the qubit.

        A K gate moves the |0> state to +|i> (positive Y) and vice versa.
        """
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.K, qubit_id=self.qubit_id
        )

    def S(self) -> None:
        """Apply an S gate on the qubit.

        An S gate is a Z-rotation with angle pi/2.
        """
        self.builder._build_cmds_single_qubit(
            instr=GenericInstr.S, qubit_id=self.qubit_id
        )

    def rot_X(
        self, n: Union[int, Template] = 0, d: int = 0, angle: Optional[float] = None
    ):
        """Do a rotation around the X-axis of the specified angle.

        The angle is interpreted as ǹ * pi / 2 ^d` radians.
        For example, (n, d) = (1, 2) represents an angle of pi/4 radians.
        If `angle` is specified, `n` and `d` are ignored and this instruction is
        automatically converted into a sequence of (n, d) rotations such that the
        discrete (n, d) values approximate the original angle.

        :param n: numerator of discrete angle specification. Can be a Template,
            in which case the subroutine containing this command should first be
            instantiated before flushing.
        :param d: denomerator of discrete angle specification
        :param angle: exact floating-point angle, defaults to None
        """
        self.builder._build_cmds_single_qubit_rotation(
            instruction=GenericInstr.ROT_X,
            virtual_qubit_id=self.qubit_id,
            n=n,
            d=d,
            angle=angle,
        )

    def rot_Y(
        self, n: Union[int, Template] = 0, d: int = 0, angle: Optional[float] = None
    ):
        """Do a rotation around the Y-axis of the specified angle.

        The angle is interpreted as ǹ * pi / 2 ^d` radians.
        For example, (n, d) = (1, 2) represents an angle of pi/4 radians.
        If `angle` is specified, `n` and `d` are ignored and this instruction is
        automatically converted into a sequence of (n, d) rotations such that the
        discrete (n, d) values approximate the original angle.

        :param n: numerator of discrete angle specification. Can be a Template,
            in which case the subroutine containing this command should first be
            instantiated before flushing.
        :param d: denomerator of discrete angle specification
        :param angle: exact floating-point angle, defaults to None
        """
        self.builder._build_cmds_single_qubit_rotation(
            instruction=GenericInstr.ROT_Y,
            virtual_qubit_id=self.qubit_id,
            n=n,
            d=d,
            angle=angle,
        )

    def rot_Z(
        self, n: Union[int, Template] = 0, d: int = 0, angle: Optional[float] = None
    ):
        """Do a rotation around the Z-axis of the specified angle.

        The angle is interpreted as ǹ * pi / 2 ^d` radians. For example, (n, d)
        = (1, 2) represents an angle of pi/4 radians. If `angle` is specified,
        `n` and `d` are ignored and this instruction is automatically converted
        into a sequence of (n, d) rotations such that the discrete (n, d) values
        approximate the original angle.

        :param n: numerator of discrete angle specification. Can be a Template,
            in which case the subroutine containing this command should first be
            instantiated before flushing.
        :param d: denomerator of discrete angle specification
        :param angle: exact floating-point angle, defaults to None
        """
        self.builder._build_cmds_single_qubit_rotation(
            instruction=GenericInstr.ROT_Z,
            virtual_qubit_id=self.qubit_id,
            n=n,
            d=d,
            angle=angle,
        )

    def cnot(self, target: Qubit) -> None:
        """Apply a CNOT gate between this qubit (control) and a target qubit.

        :param target: target qubit. Should have the same connection as this qubit.
        """
        self.builder._build_cmds_two_qubit(
            instr=GenericInstr.CNOT,
            control_qubit_id=self.qubit_id,
            target_qubit_id=target.qubit_id,
        )

    def cphase(self, target: Qubit) -> None:
        """Apply a CPHASE (CZ) gate between this qubit (control) and a target qubit.

        :param target: target qubit. Should have the same connection as this qubit.
        """
        self.builder._build_cmds_two_qubit(
            instr=GenericInstr.CPHASE,
            control_qubit_id=self.qubit_id,
            target_qubit_id=target.qubit_id,
        )

    def reset(self) -> None:
        r"""Reset the qubit to the state \|0>."""
        self.builder._build_cmds_init_qubit(qubit_id=self.qubit_id)

    def free(self) -> None:
        """
        Free the qubit and its virtual ID.

        After freeing, the underlying physical qubit can be used to store another state.
        """
        self.builder._build_cmds_qfree(qubit_id=self.qubit_id)


class FutureQubit(Qubit):
    """A Qubit that will be available in the future.

    This class is very similar to the `Future` class which is used for classical
    values.
    A `FutureQubit` acts like a `Qubit` so that all qubit operations can be applied on
    it. FutureQubits are typically the result of EPR creating requests, where they
    represent the qubits that will be available when EPR generation has finished.
    """

    def __init__(self, conn: sdkconn.BaseNetQASMConnection, future_id: Future):
        """FutureQubit constructor. Typically not used directly.

        :param conn: connection through which subroutines are sent that contain this
            qubit
        :param future_id: the virtual ID this qubit will have
        """
        self._conn: sdkconn.BaseNetQASMConnection = conn

        self.qubit_id: Future = future_id

        self._activate()

    @property
    def entanglement_info(self) -> Optional[qlink_compat.LinkLayerOKTypeK]:
        raise NotImplementedError(
            "Cannot access entanglement info of a future qubit yet"
        )

    @property
    def remote_entangled_node(self) -> Optional[str]:
        raise NotImplementedError(
            "Cannot access entanglement info of a future qubit yet"
        )



"""Command-line interface of the `netqasm` executable.

This module defines the commands that may be used when using `netqasm` as a program
on the command line, as well as handlers for these commands.
"""

import importlib
import logging
import os
import time

import click

import netqasm
from netqasm.logging.glob import get_netqasm_logger, set_log_level
from netqasm.runtime.application import (
    app_instance_from_path,
    network_cfg_from_path,
    post_function_from_path,
)
from netqasm.runtime.env import get_example_apps, init_folder, new_folder
from netqasm.runtime.settings import (
    Formalism,
    Simulator,
    set_is_using_hardware,
    set_simulator,
)
from netqasm.sdk.config import LogConfig

EXAMPLE_APPS = get_example_apps()
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
QNE_FOLDER_PATH = os.path.expanduser("~") + "/.qne"

logger = get_netqasm_logger(sub_logger="cli")
logger.setLevel(logging.INFO)
logger.propagate = False
logger.handlers = []  # Clear all handlers
formatter = logging.Formatter("{message}", style="{")
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Print info and debug warnings to the console.",
)
def cli(verbose):
    """Command line interface for managing virtual python environments."""
    if verbose:
        logger.setLevel(logging.DEBUG)


@cli.group(context_settings=CONTEXT_SETTINGS)
def qne():
    """Command line interface for managing calls to the QNE apirouter."""


###########
# version #
###########


@cli.command()
def version():
    """
    Prints the version of netqasm.
    """
    print(netqasm.__version__)


option_quiet = click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="No output printed to stdout",
)

option_app_dir = click.option(
    "--app-dir",
    type=str,
    default=None,
    help="Path to app directory. " "Defaults to CWD.",
)

option_lib_dirs = click.option(
    "--lib-dirs",
    type=str,
    default=None,
    multiple=True,
    help="Path to additional library directory.",
)

option_track_lines = click.option("--track-lines/--no-track-lines", default=True)

option_app_config_dir = click.option(
    "--app-config-dir",
    type=str,
    default=None,
    help="Explicitly choose the app config directory, " "default is `app-folder`.",
)

option_log_dir = click.option(
    "--log-dir",
    type=str,
    default=None,
    help="Explicitly choose the log directory, " "default is `app-folder/log`.",
)

option_post_func_file = click.option(
    "--post-function-file",
    type=str,
    default=None,
    help="Explicitly choose the file defining the post function.",
)

option_results_file = click.option(
    "--results-file",
    type=str,
    default=None,
    help="Explicitly choose the file where the results of a post function should be stored.",
)

option_log_level = click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="WARNING",
    help="What log-level to use (DEBUG, INFO, WARNING, ERROR, CRITICAL)."
    "Note, this affects logging to stderr, not logging instructions to file.",
)

option_log_to_files = click.option(
    "--log-to-files",
    type=bool,
    default=True,
    help="Set to false to completely disable logging to files.",
)

option_specify_host = click.option(
    "--host",
    type=str,
    default="qne-staging.quantum-inspire.com",
    help="Specify the IP and port of the api router.",
)


#########
# login #
#########


# @qne.command()
# @option_specify_host
# def login(host):
#     """Generate an API token by logging in."""
#     username = click.prompt("Username")
#     password = click.prompt("Password", hide_input=True)
#     try:
#         return _login(username, password, host)
#     except (ValueError, TypeError, AssertionError) as e:
#         logger.error(e)


# def _login(username: str, password: str, host: str = "qne-staging.quantum-inspire.com"):
#     # All the login logic except the user input.
#     if not isinstance(username, str):
#         raise TypeError("The username must be a string.")
#     elif len(username) == 0:
#         raise ValueError("The username cannot be empty.")
#     if not isinstance(password, str):
#         raise TypeError("The password must be a string.")
#     elif len(password) == 0:
#         raise ValueError("The password cannot be empty.")
#     if not isinstance(host, str):
#         raise TypeError("The host address must be a string.")
#     logger.debug(f"Logging in to {host}.")
#     jwt_req = requests.post(
#         f"http://{host}/jwt/", {"username": [username], "password": [password]}
#     )
#     if jwt_req.status_code == 401:
#         raise ValueError("Invalid credentials supplied.")
#     if jwt_req.status_code != 200:
#         raise AssertionError(
#             f"Received invalid response code ({jwt_req.status_code}) from the host."
#         )
#     try:
#         jwt_token = jwt_req.json().get("access")
#     except json.JSONDecodeError:
#         raise AssertionError("Received invalid response from the host.")
#     logger.debug("Logged in successfully.")
#     jwt_header = {"Authorization": f"JWT {jwt_token}"}
#     # NOTE: Since we get a JWT token, do we need to be cautious the token will fail also?
#     logger.debug("Generating new API token.")
#     rm_req = requests.post(
#         f"http://{host}/auth/token/destroy/",
#         data={"user": username},
#         headers=jwt_header,
#     )
#     if rm_req.status_code != 204:
#         raise AssertionError(
#             f"Received invalid response code ({rm_req.status_code}) from the host."
#         )
#     api_req = requests.post(
#         f"http://{host}/auth/token/create/", data={"user": username}, headers=jwt_header
#     )
#     if api_req.status_code != 200:
#         raise AssertionError(
#             f"Received invalid response code ({api_req.status_code}) from the host."
#         )
#     api_token = api_req.json()["access"]
#     if not os.path.exists(QNE_FOLDER_PATH):
#         os.mkdir(QNE_FOLDER_PATH)
#     with open(f"{QNE_FOLDER_PATH}/api_token", "w") as f:
#         json.dump({host: (username, api_token)}, f)
#     logger.info("Generated API token successfully.")
#     logger.debug(f"The path to the token is {QNE_FOLDER_PATH}/api_token")


# @qne.command()
# def logout():
#     try:
#         host, username, token_header = _get_token_header()
#     except FileNotFoundError:
#         logger.info("User is not logged in.")
#         return
#     rm_req = requests.post(
#         f"http://{host}/auth/token/destroy/",
#         data={"user": username},
#         headers=token_header,
#     )
#     if rm_req.status_code == 401:
#         logger.debug("API token was already invalid.")
#     elif rm_req.status_code != 204:
#         raise AssertionError(
#             f"Received invalid response code ({rm_req.status_code}) from the host."
#         )
#     logger.debug("API token successfully removed from host.")
#     os.remove(f"{QNE_FOLDER_PATH}/api_token")
#     logger.info("User logged out successfully.")


# def _get_token_header():
#     # Extract username and token header from API token file.
#     with open(f"{QNE_FOLDER_PATH}/api_token", "r") as f:
#         api_token = json.load(f)
#     assert len(api_token) == 1
#     host, (username, token_value) = list(api_token.items())[0]
#     token_header = {"Authorization": f"TOKEN {token_value}"}
#     return host, username, token_header


@cli.command()
@option_app_dir
@option_track_lines
@option_log_dir
@option_post_func_file
@option_log_level
@option_log_to_files
@click.option(
    "--network-config-file",
    type=str,
    default=None,
    help="Explicitly choose the network config file, "
    "default is `app-folder/network.yaml`.",
)
@click.option(
    "--simulator",
    type=click.Choice([sim.value for sim in Simulator]),
    default=None,
    help="Choose with simulator to use, "
    "default uses what environment variable 'NETQASM_SIMULATOR' is set to, otherwise 'netsquid'",
)
@click.option(
    "--formalism",
    type=click.Choice([f.value for f in Formalism]),
    default=Formalism.KET.value,
    help="Choose which quantum state formalism is used by the simulator. Default is 'ket'.",
)
@click.option(
    "--num", type=int, default=1, help="Number of times to run this application."
)
@click.option(
    "--sim-context/--no-sim-context",
    type=bool,
    default=True,
    help="Whether to pass an AppConfig object containing simulation context to a program's main function.",
)
@click.option(
    "--hardware",
    type=click.Choice(["generic", "nv"]),
    default="generic",
    help="What quantum hardware to use in nodes if no explicit network config is specified",
)
@click.option(
    "--timer/--no-timer",
    type=bool,
    default=False,
    help="Measure and display how much time the simulation took",
)
def simulate(
    app_dir,
    track_lines,
    log_dir,
    post_function_file,
    log_level,
    log_to_files,
    network_config_file,
    simulator,
    formalism,
    num,
    sim_context,
    hardware,
    timer,
):
    """
    Simulate an application on a simulated QNodeOS.
    """
    set_log_level(log_level)

    if simulator is None:
        simulator = os.environ.get("NETQASM_SIMULATOR", Simulator.NETSQUID.value)
    else:
        simulator = Simulator(simulator)
    formalism = Formalism(formalism)
    set_simulator(simulator=simulator)

    simulate_application = importlib.import_module(
        "netqasm.sdk.external"
    ).simulate_application
    if app_dir is None:
        app_dir = "."
    app_instance = app_instance_from_path(app_dir)
    network_cfg = network_cfg_from_path(app_dir, network_config_file)
    post_function = post_function_from_path(app_dir, post_function_file)
    if log_dir is None:
        log_dir = os.path.join(app_dir, "log")
    log_cfg = LogConfig(app_dir=app_dir, log_dir=log_dir, track_lines=track_lines)

    if timer:
        start = time.perf_counter()
    simulate_application(
        app_instance=app_instance,
        num_rounds=num,
        network_cfg=network_cfg,
        formalism=formalism,
        post_function=post_function,
        log_cfg=log_cfg,
        use_app_config=sim_context,
        enable_logging=log_to_files,
        hardware=hardware,
    )
    if timer:
        print(f"finished simulation in {round(time.perf_counter() - start, 2)} seconds")


##################
# Run on QNodeOS #
##################


@cli.command()
@option_app_dir
@option_lib_dirs
@option_track_lines
@option_app_config_dir
@option_log_dir
@option_results_file
@option_log_level
@option_log_to_files
def run(
    app_dir,
    lib_dirs,
    track_lines,
    app_config_dir,
    log_dir,
    log_level,
    log_to_files,
    results_file,
):
    """
    Execute an application on QNodeOS.
    """
    set_is_using_hardware(True)

    run_application = importlib.import_module("netqasm.sdk.external").run_application

    if app_dir is None:
        app_dir = "."
    app_instance = netqasm.runtime.application.app_instance_from_path(app_dir)

    run_application(
        app_instance=app_instance,
        instr_log_dir=log_dir,
        results_file=results_file,
    )


#######
# new #
#######
@cli.command()
@click.argument(
    "path",
    type=click.Path(exists=False),
)
@click.option(
    "--template",
    type=click.Choice(EXAMPLE_APPS),
    default="teleport",
    help="Which pre-defined app to use when creating the template (default teleport)",
)
@option_quiet
def new(path, template, quiet):
    """
    Creates a new application at PATH
    """
    if os.path.exists(path):
        raise click.BadArgumentUsage(
            f"destination `{path}` already exists\n\n"
            "Use `netqasm init` to initialize the directory"
        )
    new_folder(path, template=template, quiet=quiet)


########
# init #
########
@cli.command()
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True),
    default=".",
    help="Destination to initialize, defaults to current working directory",
)
@option_quiet
def init(path, quiet):
    """
    Initializes an existing directory with missing config files.
    Does not overwrite any existing files.
    Looks for any files `app_*.py` and creates a simple network config file
    (network.yaml) containing nodes corresponding to the application files.

    By default the current directory is used.
    Another directory can be specified with the -p/--path flag.
    """
    init_folder(path, quiet=quiet)


if __name__ == "__main__":
    cli()



import logging
from concurrent.futures import ThreadPoolExecutor
from timeit import default_timer as timer

import pytest

from netqasm.logging.glob import set_log_level
from netqasm.sdk import ThreadSocket


def execute_functions(functions):
    """Executes functions in different threads"""
    with ThreadPoolExecutor(max_workers=len(functions)) as executor:
        futures = []
        for function in functions:
            future = executor.submit(function)
            futures.append(future)

        for future in futures:
            future.result()


def test_init_error():
    with pytest.raises(ValueError):
        ThreadSocket(0, 0)


def test_connection_error():
    with pytest.raises(TimeoutError):
        ThreadSocket("alice", "bob", timeout=1)


def test_connect():
    def connect_alice():
        ThreadSocket("alice", "bob", timeout=1)

    def connect_bob():
        socket = ThreadSocket("bob", "alice", timeout=1)
        socket.wait()

    execute_functions([connect_alice, connect_bob])


def test_set_callback():
    def connect_alice():
        ThreadSocket("alice", "bob", timeout=1)

    def connect_bob():
        socket = ThreadSocket("bob", "alice", timeout=1)
        assert not socket.use_callbacks
        socket.use_callbacks = True
        assert socket.use_callbacks
        socket.wait()

    execute_functions([connect_alice, connect_bob])


def test_connection_lost():
    callbacks = [0]

    class CallbackSocket(ThreadSocket):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs, use_callbacks=True)

        def conn_lost_callback(self):
            callbacks[0] += 1

    def connect_alice():
        CallbackSocket("alice", "bob", timeout=1)

    def connect_bob():
        socket = CallbackSocket("bob", "alice", timeout=1)
        socket.wait()

    execute_functions([connect_alice, connect_bob])

    assert callbacks[0] == 1


@pytest.mark.parametrize(
    "msg",
    [
        None,
        0,
        [1, 2],
    ],
)
def test_faulty_send_type(msg):
    def connect_alice():
        socket = ThreadSocket("alice", "bob", timeout=1)
        with pytest.raises(TypeError):
            socket.send(msg)

    def connect_bob():
        socket = ThreadSocket("bob", "alice", timeout=1)
        socket.wait()

    execute_functions([connect_alice, connect_bob])


def test_faulty_send_connection():
    def connect_alice():
        ThreadSocket("alice", "bob", timeout=1)

    def connect_bob():
        socket = ThreadSocket("bob", "alice", timeout=1)
        socket.wait()
        with pytest.raises(ConnectionError):
            socket.send("")

    execute_functions([connect_alice, connect_bob])


def test_send_recv():
    msg = "hello"

    def alice():
        socket = ThreadSocket("alice", "bob")
        socket.send(msg)

    def bob():
        socket = ThreadSocket("bob", "alice")
        msg_recv = socket.recv(timeout=1)
        assert msg_recv == msg

    execute_functions([alice, bob])


def test_recv_block():
    def alice():
        socket = ThreadSocket("alice", "bob")
        socket.wait()

    def bob():
        socket = ThreadSocket("bob", "alice")
        with pytest.raises(RuntimeError):
            socket.recv(block=False)
        timeout = 1
        t_start = timer()
        with pytest.raises(TimeoutError):
            socket.recv(timeout=timeout)
        t_end = timer()
        t_elapsed = t_end - t_start
        print(t_elapsed)
        assert abs(timeout - t_elapsed) < 0.2

    execute_functions([alice, bob])


def test_ping_pong_counter():
    max_value = 10

    def alice():
        socket = ThreadSocket("alice", "bob")
        counter = 0
        counter_values = [counter]
        socket.send(str(counter))
        while counter < max_value:
            counter = int(socket.recv())
            counter += 1
            counter_values.append(counter)
            socket.send(str(counter))
        socket.wait()
        print(counter_values)
        assert counter_values == list(range(0, max_value + 2, 2))

    def bob():
        socket = ThreadSocket("bob", "alice")
        counter = 0
        counter_values = []
        while counter < max_value:
            counter = int(socket.recv())
            counter += 1
            counter_values.append(counter)
            socket.send(str(counter))
        print(counter_values)
        assert counter_values == list(range(1, max_value + 2, 2))

    execute_functions([alice, bob])


def test_ping_pong_counter_callbacks():
    max_value = 10

    class PingPongSocket(ThreadSocket):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs, use_callbacks=True)
            self.counter_values = []

        def recv_callback(self, msg):
            counter = int(msg)
            if counter > max_value:
                return
            counter += 1
            self.counter_values.append(counter)
            self.send(str(counter))

    def alice():
        socket = PingPongSocket("alice", "bob")
        socket.send("0")
        print(socket.counter_values)
        assert socket.counter_values == list(range(2, max_value + 2, 2))

    def bob():
        socket = PingPongSocket("bob", "alice")
        socket.wait()
        print(socket.counter_values)
        assert socket.counter_values == list(range(1, max_value + 2, 2))

    execute_functions([alice, bob])


if __name__ == "__main__":
    set_log_level(logging.DEBUG)
    test_connect()
    test_connection_lost()
    test_send_recv()
    test_recv_block()
    test_ping_pong_counter()
    test_ping_pong_counter_callbacks()



import math

from netqasm.examples.lib.bqc import recv_meas_outcome, send_meas_cmd, teleport_state
from netqasm.logging.glob import get_netqasm_logger
from netqasm.logging.output import get_new_app_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def main(app_config=None, num_iter=2, alpha=0, beta=0, theta1=0, theta2=0, r1=0, r2=0):
    socket = Socket("client", "server", log_config=app_config.log_config)
    epr_socket = EPRSocket("server")

    # Inputs are coefficients of pi, e.g. alpha=0.5 -> angle 0.5*pi
    alpha *= math.pi
    beta *= math.pi
    theta1 *= math.pi
    theta2 *= math.pi

    app_logger = get_new_app_logger(
        app_name=app_config.app_name, log_config=app_config.log_config
    )

    client = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    with client:
        # Remote state prepration of 2 qubits.
        # The resulting states might have a phase `pi`,
        # depending on outcome p.
        app_logger.log("Remotely preparing server's first state...")
        p1 = teleport_state(epr_socket, theta1)
        client.flush()
        app_logger.log("Remotely preparing server's second state...")
        p2 = teleport_state(epr_socket, theta2)
        client.flush()
        app_logger.log("Remote state preparation finished.")
        socket.send_silent("")

        # Convert outcomes to integers to use them in calculations below.
        p1, p2 = int(p1), int(p2)

        # Send first angle to server.
        delta1 = alpha - theta1 + (p1 + r1) * math.pi
        app_logger.log(f"Sending delta1 = {delta1}")
        send_meas_cmd(socket, delta1)
        m1 = recv_meas_outcome(socket)
        app_logger.log(f"Received m1 = {m1}")

        # Send second angle to server.
        delta2 = math.pow(-1, (m1 + r1)) * beta - theta2 + (p2 + r2) * math.pi
        app_logger.log(f"Sending delta2 = {delta2}")
        send_meas_cmd(socket, delta2)
        m2 = recv_meas_outcome(socket)
        app_logger.log(f"Received m2 = {m1}")

        return {
            "delta1": delta1 / math.pi,
            "delta2": delta2 / math.pi,
            "p1": p1,
            "p2": p2,
            "m1": m1,
            "m2": m2,
        }



"""Tools for dealing with files related to NetQASM application execution.

"""

import inspect
import os
import shutil
from datetime import datetime
from functools import wraps
from itertools import combinations
from runpy import run_path
from typing import Any, Dict

from netqasm.examples import apps
from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.settings import Simulator, set_simulator
from netqasm.util.yaml import dump_yaml, load_yaml

EXAMPLE_APPS_DIR = os.path.dirname(os.path.abspath(apps.__file__))

IGNORED_FILES = [
    "__init__.py",
    "__pycache__",
    "log",
    "cysignals_crash_logs",
]

logger = get_netqasm_logger()


def load_app_config_file(app_dir, app_name) -> Any:
    ext = ".yaml"
    file_path = os.path.join(app_dir, f"{app_name}{ext}")
    if os.path.exists(file_path):
        config = load_yaml(file_path=file_path)
    else:
        config = None
    if config is None:
        return {}
    else:
        return config


def get_roles_config_path(app_dir):
    ext = ".yaml"
    file_path = os.path.join(app_dir, f"roles{ext}")
    return file_path


def load_roles_config(roles_config_file):
    if os.path.exists(roles_config_file):
        return load_yaml(roles_config_file)
    else:
        return None


def load_app_files(app_dir) -> Dict[str, str]:  # app_name -> file
    app_tag = "app_"
    ext = ".py"
    app_files = {}
    for entry in os.listdir(app_dir):
        if entry.startswith(app_tag) and entry.endswith(".py"):
            app_name = entry[len(app_tag) : -len(ext)]
            app_files[app_name] = entry
    if len(app_files) == 0:
        raise ValueError(
            f"directory {app_dir} does not seem to be a application directory (no app_xxx.py files)"
        )
    return app_files


def get_log_dir(app_dir):
    log_dir = os.path.join(app_dir, "log")
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)
    return log_dir


def get_timed_log_dir(log_dir):
    now = datetime.now().strftime("%Y%m%d-%H%M%S")
    timed_log_dir = os.path.join(log_dir, now)
    if not os.path.exists(timed_log_dir):
        os.mkdir(timed_log_dir)
    return timed_log_dir


def get_post_function_path(app_dir):
    return os.path.join(app_dir, "post_function.py")


def load_post_function(post_function_file):
    if not os.path.exists(post_function_file):
        return None
    return run_path(post_function_file)["main"]


def get_results_path(timed_log_dir):
    return os.path.join(timed_log_dir, "results.yaml")


def new_folder(path, template="teleport", quiet=False):
    """Used by the CLI to create an app folder template

    Parameters
    ----------
    path : str
        Path to the directory
    template : str
        Which pre-defined app to use as template
    quiet : bool
        Whether to print info to stdout or not (default `False`)
    """
    assert not os.path.exists(path), "Destination already exists"
    os.mkdir(path)
    template_example_dir = os.path.join(EXAMPLE_APPS_DIR, template)
    for entry in os.listdir(template_example_dir):
        entry_path = os.path.join(template_example_dir, entry)
        if entry not in IGNORED_FILES:
            target_path = os.path.join(path, entry)
            if os.path.isfile(entry_path):
                shutil.copyfile(entry_path, target_path)
            elif os.path.isdir(entry_path):
                shutil.copytree(entry_path, target_path)
    if not quiet:
        print(f"Creating application template ({template} example) in `{path}`")


def init_folder(path, quiet=False):
    """Used by the CLI to initialize a directory by adding missing config files.

    Parameters
    ----------
    path : str
        Path to the directory
    quiet : bool
        Whether to print info to stdout or not (default `False`)
    """
    app_files = load_app_files(path)
    file_added = False

    # Create network file if non-existant
    network_file_path = os.path.join(path, "network.yaml")
    if not os.path.exists(network_file_path):
        _create_new_network_file(
            app_files=app_files,
            file_path=network_file_path,
            quiet=quiet,
        )
        file_added = True

    # Create roles file if non-existant
    roles_file_path = os.path.join(path, "roles.yaml")
    if not os.path.exists(roles_file_path):
        _create_new_roles_file(
            app_files=app_files,
            file_path=roles_file_path,
            quiet=quiet,
        )
        file_added = True

    # Create input files if non-existant
    for app_name, app_file in app_files.items():
        input_file_path = os.path.join(path, f"{app_name}.yaml")
        if not os.path.exists(input_file_path):
            app_file_path = os.path.join(path, app_file)
            _create_new_input_file(
                app_name=app_name,
                app_file_path=app_file_path,
                file_path=input_file_path,
                quiet=quiet,
            )
            file_added = True

    # Create README.md if non-existant
    readme_file_path = os.path.join(path, "README.md")
    if not os.path.exists(readme_file_path):
        _create_new_readme_file(
            file_path=readme_file_path,
        )
        file_added = True

    # Create results_config.json if non-existant
    results_config_file_path = os.path.join(path, "results_config.json")
    if not os.path.exists(results_config_file_path):
        _create_new_results_config_file(
            file_path=results_config_file_path,
        )
        file_added = True

    if file_added:
        if not quiet:
            if path == ".":
                path_str = "current path"
            else:
                path_str = "`{path}`"
            print(f"Initialized {path_str} with missing config files")
    else:
        if not quiet:
            print("No files needed to be added")


def file_creation_notify(func):
    """Decorator for notification about file creation"""

    @wraps(func)
    def new_func(file_path, *args, quiet=False, **kwargs):
        func(file_path, *args, quiet=quiet, **kwargs)

        if not quiet:
            print(f"Created file `{file_path}`")

    return new_func


@file_creation_notify
def _create_new_network_file(file_path, app_files, quiet=False):
    # Create nodes
    nodes = []
    for app_name in app_files.keys():
        qubit = {
            "id": 0,
            "t1": 0,
            "t2": 0,
        }
        node = {
            "name": app_name,
            "gate_fidelity": 1.0,
            "qubits": [qubit],
        }
        nodes.append(node)

    # Create links
    links = []
    for i, (app_name1, app_name2) in enumerate(combinations(app_files.keys(), 2)):
        if app_name1 == app_name2:
            continue
        link = {
            "name": f"ch{i}",
            "node_name1": app_name1,
            "node_name2": app_name2,
            "noise_type": "Depolarise",
            "fidelity": 1.0,
        }
        links.append(link)

    # Create network
    network = {
        "nodes": nodes,
        "links": links,
    }

    dump_yaml(data=network, file_path=file_path)


@file_creation_notify
def _create_new_roles_file(file_path, app_files, quiet=False):
    # Create roles
    roles = {}
    for app_name in app_files.keys():
        roles[app_name] = app_name

    dump_yaml(data=roles, file_path=file_path)


@file_creation_notify
def _create_new_input_file(file_path, app_name, app_file_path, quiet=False):
    arguments = _find_argument_for_app_file(app_file_path)

    dump_yaml(data=arguments, file_path=file_path)


def _find_argument_for_app_file(app_file_path):
    set_simulator(Simulator.DEBUG)
    members = run_path(app_file_path)
    main = members.get("main")
    if main is None:
        raise RuntimeError(f"File `{app_file_path}` does not have a `main`-function")

    signature = inspect.signature(main)
    return {
        param.name: param.default
        for param in signature.parameters.values()
        if param.name != "app_config"
    }


@file_creation_notify
def _create_new_readme_file(file_path, quiet=False):
    with open(file_path, "w") as f:
        f.write(
            "# Application name\n"
            "Some description of the application.\n"
            "\n"
            "## Inputs\n"
            "Description of inputs.\n"
            "\n"
            "## Outputs\n"
            "Description of outputs.\n"
        )


@file_creation_notify
def _create_new_results_config_file(file_path, quiet=False):
    with open(file_path, "w") as f:
        f.write(
            r"""[
    [
        {
            "output_type": "text",
            "title": "Results",
            "parameters": {
                "content": "Information about the results."
            }
        }
    ]
]"""
        )


def get_example_apps():
    return [
        app_name
        for app_name in os.listdir(EXAMPLE_APPS_DIR)
        if app_name not in IGNORED_FILES
    ]



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()
inner_num = 10
outer_num = 8
inner_reg = "R0"
outer_reg = "R1"


def run_alice():
    with NetQASMConnection("Alice") as alice:

        array = alice.new_array(2, init_values=[0, 0])
        i = array.get_future_index(0)
        j = array.get_future_index(1)

        def outer_body(alice, _):
            def inner_body(alice, _):
                q = Qubit(alice)
                q.free()
                j.add(1)

            i.add(1)

            alice.loop_body(inner_body, inner_num, loop_register=inner_reg)

        alice.loop_body(outer_body, outer_num, loop_register=outer_reg)
    assert i == outer_num
    assert j == outer_num * inner_num


def test_nested_loop():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)


if __name__ == "__main__":
    test_nested_loop()



import importlib
import random

import numpy as np
import pytest

from netqasm.examples.apps.blind_grover.app_client import main as blind_grover_client
from netqasm.examples.apps.blind_grover.app_server import main as blind_grover_server
from netqasm.examples.apps.blind_rotation.app_client import (
    main as blind_rotation_client,
)
from netqasm.examples.apps.blind_rotation.app_server import (
    main as blind_rotation_server,
)
from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk.external import simulate_application

if get_simulator() == Simulator.NETSQUID:
    from netsquid.qubits import qubitapi

logger = get_netqasm_logger()


def fidelity_ok(qubit, dm, threshold=0.999):
    fidelity = qubitapi.fidelity(qubit, dm)
    return fidelity > threshold


def run_blind_rotation():
    ns = importlib.import_module("netsquid")
    num_iter = 4
    num_qubits = num_iter + 1
    phi = [random.uniform(0, 2 * np.pi) for _ in range(num_iter)]
    theta = [random.uniform(0, 2 * np.pi) for _ in range(num_qubits)]
    r = [random.randint(0, 1) for _ in range(num_iter)]

    alice_app_inputs = {"num_iter": num_iter, "theta": theta, "phi": phi, "r": r}

    bob_app_inputs = {"num_iter": num_iter}

    app_instance = default_app_instance(
        [
            ("client", blind_rotation_client),
            ("server", blind_rotation_server),
        ]
    )
    app_instance.program_inputs["client"] = alice_app_inputs
    app_instance.program_inputs["server"] = bob_app_inputs

    results = simulate_application(app_instance, enable_logging=False)[0]

    output_state = results["app_server"]["output_state"]
    s = results["app_client"]["s"]
    m = results["app_client"]["m"]
    r = results["app_client"]["r"]
    theta = results["app_client"]["theta"]

    s.extend([0, 0])
    m.extend([0, 0])
    r.extend([0, 0])

    # output should be (n = num_iter):
    # Rz(theta[n]) Z^m[n] X^{s[n-1]^r[n-1]} Z^{s[n-2]^r[n-2]} H Rz(phi[n-1]) H Rz(phi[n-2] ... |+>
    ref = ns.qubits.create_qubits(1)[0]
    ns.qubits.operate(ref, ns.H)
    for i in range(num_iter):
        ns.qubits.operate(ref, ns.qubits.create_rotation_op(phi[i], (0, 0, 1)))
        ns.qubits.operate(ref, ns.H)
    if m[num_iter] == 1:
        ns.qubits.operate(ref, ns.Z)
    if (s[num_iter - 1] ^ r[num_iter - 1]) == 1:
        ns.qubits.operate(ref, ns.X)
    if (s[num_iter - 2] ^ r[num_iter - 2]) == 1:
        ns.qubits.operate(ref, ns.Z)
    ns.qubits.operate(ref, ns.qubits.create_rotation_op(theta[num_iter], (0, 0, 1)))

    assert fidelity_ok(ref, np.array(output_state))


@pytest.mark.skipif(
    get_simulator() != Simulator.NETSQUID,
    reason="Can only run blind rotation tests using netsquid for now",
)
def test_blind_rotation():
    num = 3
    for _ in range(num):
        run_blind_rotation()


def run_blind_grover():
    b0 = random.randint(0, 1)
    b1 = random.randint(0, 1)

    r1 = random.randint(0, 1)
    r2 = random.randint(0, 1)

    theta1 = random.uniform(0, 2 * np.pi)
    theta2 = random.uniform(0, 2 * np.pi)

    alice_app_inputs = {
        "b0": b0,
        "b1": b1,
        "r1": r1,
        "r2": r2,
        "theta1": theta1,
        "theta2": theta2,
    }

    bob_app_inputs = {}

    app_instance = default_app_instance(
        [
            ("client", blind_grover_client),
            ("server", blind_grover_server),
        ]
    )
    app_instance.program_inputs["client"] = alice_app_inputs
    app_instance.program_inputs["server"] = bob_app_inputs

    # results = run_applications(applications)
    results = simulate_application(app_instance, enable_logging=False)[0]

    m0 = results["app_client"]["result0"]
    m1 = results["app_client"]["result1"]

    assert b0 == m0
    assert b1 == m1


@pytest.mark.skipif(
    get_simulator() != Simulator.NETSQUID,
    reason="Can only run blind rotation tests using netsquid for now",
)
def test_blind_grover():
    num = 3
    for _ in range(num):
        run_blind_grover()


if __name__ == "__main__":
    run_blind_rotation()
    run_blind_grover()



import ctypes
from enum import Enum

############
# METADATA #
############
NETQASM_VERSION = ctypes.c_uint8 * 2
APP_ID = ctypes.c_uint16


class Metadata(ctypes.Structure):
    _fields_ = [
        ("netqasm_version", NETQASM_VERSION),
        ("app_id", APP_ID),
    ]


METADATA_BYTES = len(bytes(Metadata()))


########
# BODY #
########
INTEGER = ctypes.c_int32
INTEGER_BITS = len(bytes(INTEGER())) * 8  # type: ignore
IMMEDIATE = ctypes.c_uint8
IMMEDIATE_BITS = len(bytes(IMMEDIATE())) * 8  # type: ignore

ADDRESS = INTEGER
ADDRESS_BITS: int = len(bytes(ADDRESS())) * 8  # type: ignore

INSTR_ID = ctypes.c_uint8

REG_TYPE = ctypes.c_uint8
REG_BITS = len(bytes(REG_TYPE())) * 8  # type: ignore
# Num bits in register name
REG_NAME_BITS = 2
# Num bits in register index
REG_INDEX_BITS = 4

COMMAND_BYTES = 7

PADDING_FIELD = "padding"


class OptionalInt(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_uint8),
        ("value", INTEGER),
    ]

    _NULL_TYPE = 0x00
    _INT_TYPE = 0x01

    def __init__(self, value):
        if value is None:
            self.type = self._NULL_TYPE
            self.value = 0
        else:
            self.type = self._INT_TYPE
            self.value = value

    def value(self):
        if self.type == self._NULL_TYPE:
            return None
        elif self.type == self._INT_TYPE:
            return self.value
        else:
            raise TypeError(f"Unknown type {self.type}")


class RegisterName(Enum):
    # Standard register
    R = 0
    # Register for constants
    C = 1
    # Qubit addresses
    Q = 2
    # Measurment outcomes
    M = 3


class Register(ctypes.Structure):
    _fields_ = [
        ("register_name", REG_TYPE, REG_NAME_BITS),
        ("register_index", REG_TYPE, REG_INDEX_BITS),
        (PADDING_FIELD, REG_TYPE, REG_BITS - REG_NAME_BITS - REG_INDEX_BITS),
    ]


class Address(ctypes.Structure):
    _fields_ = [
        ("address", ADDRESS),
    ]


class ArrayEntry(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("address", Address),
        ("index", Register),
    ]


class ArraySlice(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("address", Address),
        ("start", Register),
        ("stop", Register),
    ]


class Command(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("id", INSTR_ID),
    ]

    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except TypeError as err:
            raise TypeError(
                f"command {self.__class__.__name__} could not be created, since: {err}"
            )


def add_padding(fields):
    """Used to add correct amount of padding for commands to make them fixed-length"""
    # TODO better way?
    class TmpCommand(Command):
        pass

    TmpCommand._fields_ = fields
    current_num_bytes = len(bytes(TmpCommand()))
    total_num_bytes = COMMAND_BYTES
    pad_num_bytes = total_num_bytes - current_num_bytes
    assert pad_num_bytes >= 0
    new_fields = fields + [(PADDING_FIELD, ctypes.c_uint8 * pad_num_bytes)]
    return new_fields


class NoOperandCommand(Command):
    _fields_ = add_padding([])


class RegCommand(Command):
    _fields_ = add_padding(
        [
            ("reg", Register),
        ]
    )


class RegRegCommand(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
        ]
    )


class MeasCommand(Command):
    _fields_ = add_padding(
        [
            ("qubit", Register),
            ("outcome", Register),
        ]
    )


class RegImmImmCommand(Command):
    _fields_ = add_padding(
        [
            ("reg", Register),
            ("imm0", IMMEDIATE),
            ("imm1", IMMEDIATE),
        ]
    )


class RegRegImmImmCommand(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("imm0", IMMEDIATE),
            ("imm1", IMMEDIATE),
        ]
    )


class RegRegImm4Command(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("imm0", IMMEDIATE),
            ("imm1", IMMEDIATE),
            ("imm2", IMMEDIATE),
            ("imm3", IMMEDIATE),
        ]
    )


class RegRegRegCommand(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("reg2", Register),
        ]
    )


class RegRegRegRegCommand(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("reg2", Register),
            ("reg3", Register),
        ]
    )


class ImmCommand(Command):
    _fields_ = add_padding(
        [
            ("imm", INTEGER),
        ]
    )


class ImmImmCommand(Command):
    _fields_ = add_padding(
        [
            ("imm0", IMMEDIATE),
            ("imm1", IMMEDIATE),
        ]
    )


class RegRegImmCommand(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("imm", INTEGER),
        ]
    )


class RegImmCommand(Command):
    _fields_ = add_padding(
        [
            ("reg", Register),
            ("imm", INTEGER),
        ]
    )


class RegEntryCommand(Command):
    _fields_ = add_padding(
        [
            ("reg", Register),
            ("entry", ArrayEntry),
        ]
    )


class RegAddrCommand(Command):
    _fields_ = add_padding(
        [
            ("reg", Register),
            ("addr", Address),
        ]
    )


class ArrayEntryCommand(Command):
    _fields_ = add_padding(
        [
            ("entry", ArrayEntry),
        ]
    )


class ArraySliceCommand(Command):
    _fields_ = add_padding(
        [
            ("slice", ArraySlice),
        ]
    )


class SingleRegisterCommand(Command):
    _fields_ = add_padding(
        [
            ("register", Register),
        ]
    )


class ArrayCommand(Command):
    _fields_ = add_padding(
        [
            ("size", Register),
            ("address", Address),
        ]
    )


class AddrCommand(Command):
    _fields_ = add_padding(
        [
            ("addr", Address),
        ]
    )


class Reg5Command(Command):
    _fields_ = add_padding(
        [
            ("reg0", Register),
            ("reg1", Register),
            ("reg2", Register),
            ("reg3", Register),
            ("reg4", Register),
        ]
    )


class RecvEPRCommand(Command):
    _fields_ = add_padding(
        [
            ("remote_node_id", Register),
            ("epr_socket_id", Register),
            ("qubit_address_array", Register),
            ("ent_results_array", Register),
        ]
    )


COMMANDS = [
    RegCommand,
    RegRegCommand,
    MeasCommand,
    RegImmImmCommand,
    RegRegImmImmCommand,
    RegRegImm4Command,
    RegRegRegCommand,
    ImmCommand,
    ImmImmCommand,
    NoOperandCommand,
    RegImmCommand,
    RegRegImmCommand,
    RegEntryCommand,
    ArrayEntryCommand,
    ArraySliceCommand,
    SingleRegisterCommand,
    ArrayCommand,
    Reg5Command,
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .futures import BaseFuture


class SdkConstraint:
    """A constraint that can be used in exit conditions in SDK loops."""

    pass


class ValueAtMostConstraint(SdkConstraint):
    """A constraint that a certain variable should be at most another value."""

    def __init__(self, future: BaseFuture, value: int) -> None:
        """ValueAtMostConstraint constructor.

        :param future: the variable that should be at most the given value
        :param value: the maximum value that the given future may have
        """
        self._future = future
        self._value = value

    @property
    def future(self) -> BaseFuture:
        return self._future

    @property
    def value(self) -> int:
        return self._value



import abc


class NetworkInfo:
    """Global information about the current quantum network environment.

    This class is a container for static functions that provide information about
    the current network setting.
    Applications may use this information to e.g. obtain node IDs or map party names
    to nodes.

    Concrete runtime contexts (like a simulator, or a real hardware setup) should
    override these methods to provide the information specific to that context.
    """

    @classmethod
    @abc.abstractmethod
    def _get_node_id(cls, node_name: str) -> int:
        """Returns the node id for the node with the given name"""
        # Should be subclassed
        pass

    @classmethod
    @abc.abstractmethod
    def _get_node_name(cls, node_id: int) -> str:
        """Returns the node name for the node with the given ID"""
        # Should be subclassed
        pass

    @classmethod
    @abc.abstractmethod
    def get_node_id_for_app(cls, app_name: str) -> int:
        """Returns the node id for the app with the given name"""
        # Should be subclassed
        pass

    @classmethod
    @abc.abstractmethod
    def get_node_name_for_app(cls, app_name: str) -> str:
        """Returns the node name for the app with the given name"""
        # Should be subclassed
        pass



from src.protocol import anonymous_transmission


def main(
    app_config=None,
    sender=False,
    receiver=False,
    phi=0.0,
    theta=0.0,
):

    return anonymous_transmission(
        app_name="alice",
        app_config=app_config,
        sender=sender,
        receiver=receiver,
        phi=phi,
        theta=theta,
    )


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()

# TODO depending on how post functions are implemented in simulaqron
# this should not be a variable outside `run_alice` since it won't be
# affected when running that function in a seperate process.
# We should instead use the results returned from `run_applications`.
outcomes = []


def run_alice():
    epr_socket = EPRSocket("Bob")
    with NetQASMConnection("Alice", epr_sockets=[epr_socket]) as alice:
        # Create a qubit
        q = Qubit(alice)
        q.H()

        # Create entanglement
        epr = epr_socket.create()[0]

        # Teleport
        q.cnot(epr)
        q.H()
        m1 = q.measure()
        m2 = epr.measure()
        outcomes.append(m1)
        outcomes.append(m2)


def run_bob():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]):
        epr_socket.recv()


def post_function(backend):
    m1, m2 = outcomes
    logger.info(f"m1, m2 = {m1}, {m2}")
    expected_states = {
        (0, 0): np.array([[0.5, 0.5], [0.5, 0.5]]),
        (0, 1): np.array([[0.5, 0.5], [0.5, 0.5]]),
        (1, 0): np.array([[0.5, -0.5], [-0.5, 0.5]]),
        (1, 1): np.array([[0.5, -0.5], [-0.5, 0.5]]),
    }
    state = backend.nodes["Bob"].qmemory._get_qubits(0)[0].qstate.qrepr.reduced_dm()
    logger.info(f"state = {state}")
    expected = expected_states[m1, m2]
    logger.info(f"expected = {expected}")
    assert np.all(np.isclose(expected, state))


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_teleport_without_corrections():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )



"""EPR Socket interface."""

from __future__ import annotations

import abc
import logging
from contextlib import contextmanager
from typing import TYPE_CHECKING, Callable, ContextManager, List, Optional, Tuple, Union

from netqasm.logging.glob import get_netqasm_logger
from netqasm.qlink_compat import (
    EPRRole,
    EPRType,
    LinkLayerOKTypeK,
    LinkLayerOKTypeM,
    LinkLayerOKTypeR,
    RandomBasis,
    TimeUnit,
)
from netqasm.sdk.build_epr import EprMeasBasis, basis_to_rotation
from netqasm.sdk.builder import EntRequestParams, EprKeepResult, EprMeasureResult
from netqasm.sdk.futures import RegFuture

from .qubit import FutureQubit, Qubit

if TYPE_CHECKING:
    from netqasm.sdk import connection

T_LinkLayerOkList = Union[
    List[LinkLayerOKTypeK], List[LinkLayerOKTypeM], List[LinkLayerOKTypeR]
]


class EPRSocket(abc.ABC):
    """EPR socket class. Used to generate entanglement with a remote node.

    An EPR socket represents a connection with a single remote node through which
    EPR pairs can be generated. Its main interfaces are the `create` and `recv`
    methods. A typical use case for two nodes is that they both create an EPR socket
    to the other node, and during the protocol, one of the nodes does `create`
    operations on its socket while the other node does `recv` operations.

    A `create` operation asks the network stack to initiate generation of EPR pairs
    with the remote node. Depending on the type of generation, the result of this
    operation can be qubit objects or measurement outcomes.
    A `recv` operation asks the network stack to wait for the remote node to initiate
    generation of EPR pairs. Again, the result can be qubit objects or measurement
    outcomes.

    Each `create` operation on one node must be matched by a `recv` operation on the
    other node. Since "creating" and "receiving" must happen at the same time, a node
    that is doing a `create` operation on its socket cannot advance until the other
    node does the corresponding `recv`. This is different from classical network
    sockets where a "send" operation (roughly anologous to `create` in an EPR socket)
    does not block on the remote node receiving it.

    An EPR socket is identified by a triple consisting of (1) the remote node ID,
    (2) the local socket ID and (3) the remote socket ID.
    Two nodes that want to generate EPR pairs with each other should make sure that the
    IDs in their local sockets match.
    """

    def __init__(
        self,
        remote_app_name: str,
        epr_socket_id: int = 0,
        remote_epr_socket_id: int = 0,
        min_fidelity: int = 100,
    ):
        """Create an EPR socket. It still needs to be registered with the network
        stack separately.

        Registering and opening the EPR socket is currently done automatically by the
        connection that uses this EPR socket, specifically when a context is opened
        with that connection.

        :param remote_app_name: name of the remote party (i.e. the role, like "client",
            not necessarily the node name like "delft")
        :param epr_socket_id: local socket ID, defaults to 0
        :param remote_epr_socket_id: remote socket ID, defaults to 0. Note that this
            must match with the local socket ID of the remote node's EPR socket.
        :param min_fidelity: minimum desired fidelity for EPR pairs generated over this
            socket, in percentages (i.e. range 0-100). Defaults to 100.
        """
        self._conn: Optional[connection.BaseNetQASMConnection] = None
        self._remote_app_name: str = remote_app_name
        self._remote_node_id: Optional[
            int
        ] = None  # Gets set when the connection is set
        self._epr_socket_id: int = epr_socket_id
        self._remote_epr_socket_id: int = remote_epr_socket_id

        if (
            not isinstance(min_fidelity, int)
            or (min_fidelity < 0)
            or min_fidelity > 100
        ):
            raise ValueError(
                f"min_fidelity must be an integer in the range [0, 100], not {min_fidelity}"
            )
        self._min_fidelity: int = min_fidelity

        self._logger: logging.Logger = get_netqasm_logger(
            f"{self.__class__.__name__}({self._remote_app_name}, {self._epr_socket_id})"
        )

    @property
    def conn(self) -> connection.BaseNetQASMConnection:
        """Get the underlying :class:`NetQASMConnection`"""
        if self._conn is None:
            raise RuntimeError("EPRSocket does not have an open connection")
        return self._conn

    @conn.setter
    def conn(self, conn: connection.BaseNetQASMConnection):
        self._conn = conn
        self._remote_node_id = self._get_node_id(app_name=self._remote_app_name)

    @property
    def remote_app_name(self) -> str:
        """Get the remote application name"""
        return self._remote_app_name

    @property
    def remote_node_id(self) -> int:
        """Get the remote node ID"""
        if self._remote_node_id is None:
            raise RuntimeError("Remote Node ID has not been initialized")
        return self._remote_node_id

    @property
    def epr_socket_id(self) -> int:
        """Get the EPR socket ID"""
        return self._epr_socket_id

    @property
    def remote_epr_socket_id(self) -> int:
        """Get the remote EPR socket ID"""
        return self._remote_epr_socket_id

    @property
    def min_fidelity(self) -> int:
        """Get the desired minimum fidelity"""
        return self._min_fidelity

    def create_keep(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> List[Qubit]:
        """Ask the network stack to generate EPR pairs with the remote node and keep
        them in memory.

        A `create_keep` operation must always be matched by a `recv_keep` operation on
        the remote node.

        If `sequential` is False (default), this operation returns a list of Qubit
        objects representing the local qubits that are each one half of the generated
        pairs. These qubits can then be manipulated locally just like locally
        initialized qubits, by e.g. applying gates or measuring them.
        Each qubit also contains information about the entanglement generation that
        lead to its creation, and can be accessed by its `entanglement_info` property.

        A typical example for just generating one pair with another node would be:

        .. code-block::

            q = epr_socket.create_keep()[0]
            # `q` can now be used as a normal qubit

        If `sequential` is False (default), the all requested EPR pairs are generated
        at once, before returning the results (qubits or entanglement info objects).

        If `sequential` is True, a callback function (`post_routine`) should be
        specified. After generating one EPR pair, this callback will be called, before
        generating the next pair. This method can e.g. be used to generate many EPR
        pairs (more than the number of physical qubits available), by measuring (and
        freeing up) each qubit before the next pair is generated.

        For example:

        .. code-block::

            outcomes = alice.new_array(num)

            def post_create(conn, q, pair):
                q.H()
                outcome = outcomes.get_future_index(pair)
                q.measure(outcome)
            epr_socket.create_keep(number=num, post_routine=post_create, sequential=True)


        :param number: number of EPR pairs to generate, defaults to 1
        :param post_routine: callback function for each genated pair. Only used if
            `sequential` is True.
            The callback should take three arguments `(conn, q, pair)` where
            * `conn` is the connection (e.g. `self`)
            * `q` is the entangled qubit (of type `FutureQubit`)
            * `pair` is a register holding which pair is handled (0, 1, ...)
        :param sequential: whether to use callbacks after each pair, defaults to False
        :param time_unit: which time unit to use for the `max_time` parameter
        :param max_time: maximum number of time units (see `time_unit`) the Host is
            willing to wait for entanglement generation of a single pair. If generation
            does not succeed within this time, the whole subroutine that this request
            is part of is reset and run again by the quantum node controller.
        :param min_fidelity_all_at_end: the minimum fidelity that *all* entangled
            qubits should ideally still have at the moment the last qubit has been
            generated. For example, when specifying `number=2` and
            `min_fidelity_all_at_end=80`, the the program will automatically try to
            make sure that both qubits have a fidelity of at least 80% when the
            second qubit has been generated. It will attempt to do this by
            automatically re-trying the entanglement generation if the fidelity
            constraint is not satisfied. This is however an *attempt*, and not
            a guarantee!.
        :param max_tries: maximum number of re-tries should be made to try and achieve
            the `min_fidelity_all_at_end` constraint.
        :return: list of qubits created
        """

        qubits, _ = self.conn.builder.sdk_create_epr_keep(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=post_routine,
                sequential=sequential,
                time_unit=time_unit,
                max_time=max_time,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            ),
        )
        return qubits

    def create_keep_with_info(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
        min_fidelity_all_at_end: Optional[int] = None,
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Same as create_keep but also return the EPR generation information coming
        from the network stack.

        For more information see the documentation of `create_keep`.

        :param number: number of pairs to generate, defaults to 1
        :return: tuple with (1) list of qubits created, (2) list of EprKeepResult objects
        """
        qubits, info = self.conn._builder.sdk_create_epr_keep(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=post_routine,
                sequential=sequential,
                time_unit=time_unit,
                max_time=max_time,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
            ),
        )
        return qubits, info

    def create_measure(
        self,
        number: int = 1,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
        basis_local: EprMeasBasis = None,
        basis_remote: EprMeasBasis = None,
        rotations_local: Tuple[int, int, int] = (0, 0, 0),
        rotations_remote: Tuple[int, int, int] = (0, 0, 0),
        random_basis_local: Optional[RandomBasis] = None,
        random_basis_remote: Optional[RandomBasis] = None,
    ) -> List[EprMeasureResult]:
        """Ask the network stack to generate EPR pairs with the remote node and
        measure them immediately (on both nodes).

        A `create_measure` operation must always be matched by a `recv_measure`
        operation on the remote node.

        This operation returns a list of Linklayer response objects. These objects
        contain information about the entanglement generation and includes the
        measurement outcome and basis used. Note that all values are `Future` objects.
        This means that the current subroutine must be flushed before the values
        become defined.

        An example for generating 10 pairs with another node that are immediately
        measured:

        .. code-block::

            # list of Futures that become defined when subroutine is flushed
            outcomes = []
            with NetQASMConnection("alice", epr_sockets=[epr_socket]):
                ent_infos = epr_socket.create(number=10, tp=EPRType.M)
                for ent_info in ent_infos:
                    outcomes.append(ent_info.measurement_outcome)

        The basis to measure in can also be specified. There are 3 ways to specify a
        basis:

        * using one of the `EprMeasBasis` variants
        * by specifying 3 rotation angles, interpreted as an X-rotation, a Y-rotation
          and another X-rotation. For example, setting `rotations_local` to (8, 0, 0)
          means that before measuring, an X-rotation of 8*pi/16 = pi/2 radians is
          applied to the qubit.
        * using one of the `RandomBasis` variants, in which case one of the bases of
          that variant is chosen at random just before measuring

        NOTE: the node that initiates the entanglement generation, i.e. the one that
        calls `create` on its EPR socket, also controls the measurement bases of the
        receiving node (by setting e.g. `rotations_remote`). The receiving node cannot
        change this.

        :param number: number of EPR pairs to generate, defaults to 1
        :param time_unit: which time unit to use for the `max_time` parameter
        :param max_time: maximum number of time units (see `time_unit`) the Host is
            willing to wait for entanglement generation of a single pair. If generation
            does not succeed within this time, the whole subroutine that this request
            is part of is reset and run again by the quantum node controller.
        :param basis_local: basis to measure in on this node for M-type requests
        :param basis_remote: basis to measure in on the remote node for M-type requests
        :param rotations_local: rotations to apply before measuring on this node
        :param rotations_remote: rotations to apply before measuring on remote node
        :param random_basis_local: random bases to choose from when measuring on this
            node
        :param random_basis_remote: random bases to choose from when measuring on
            the remote node
        :return: list of entanglement info objects per created pair.
        """

        if basis_local is not None:
            rotations_local = basis_to_rotation(basis_local)
        if basis_remote is not None:
            rotations_remote = basis_to_rotation(basis_remote)

        return self.conn.builder.sdk_create_epr_measure(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=None,
                sequential=False,
                time_unit=time_unit,
                max_time=max_time,
                random_basis_local=random_basis_local,
                random_basis_remote=random_basis_remote,
                rotations_local=rotations_local,
                rotations_remote=rotations_remote,
            ),
        )

    def create_rsp(
        self,
        number: int = 1,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
        basis_local: EprMeasBasis = None,
        rotations_local: Tuple[int, int, int] = (0, 0, 0),
        random_basis_local: Optional[RandomBasis] = None,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> List[EprMeasureResult]:
        """Ask the network stack to do remote preparation with the remote node.

        A `create_rsp` operation must always be matched by a `recv_epr` operation
        on the remote node.

        This operation returns a list of Linklayer response objects. These objects
        contain information about the entanglement generation and includes the
        measurement outcome and basis used. Note that all values are `Future` objects.
        This means that the current subroutine must be flushed before the values
        become defined.

        An example for generating 10 pairs with another node that are immediately
        measured:

        .. code-block::

            m: LinkLayerOKTypeM = epr_socket.create_rsp(tp=EPRType.R)[0]
            print(m.measurement_outcome)
            # remote node now has a prepared qubit

        The basis to measure in can also be specified.
        There are 3 ways to specify a basis:

        * using one of the `EprMeasBasis` variants
        * by specifying 3 rotation angles, interpreted as an X-rotation, a Y-rotation
          and another X-rotation. For example, setting `rotations_local` to (8, 0, 0)
          means that before measuring, an X-rotation of 8*pi/16 = pi/2 radians is
          applied to the qubit.
        * using one of the `RandomBasis` variants, in which case one of the bases of
          that variant is chosen at random just before measuring

        :param number: number of EPR pairs to generate, defaults to 1
        :param time_unit: which time unit to use for the `max_time` parameter
        :param max_time: maximum number of time units (see `time_unit`) the Host is
            willing to wait for entanglement generation of a single pair. If generation
            does not succeed within this time, the whole subroutine that this request
            is part of is reset and run again by the quantum node controller.
        :param basis_local: basis to measure in on this node for M-type requests
        :param rotations_local: rotations to apply before measuring on this node
        :param random_basis_local: random bases to choose from when measuring on this
            node
        :param min_fidelity_all_at_end: the minimum fidelity that *all* entangled
            qubits should ideally still have at the moment the last qubit has been
            generated. For example, when specifying `number=2` and
            `min_fidelity_all_at_end=80`, the the program will automatically try to
            make sure that both qubits have a fidelity of at least 80% when the
            second qubit has been generated. It will attempt to do this by
            automatically re-trying the entanglement generation if the fidelity
            constraint is not satisfied. This is however an *attempt*, and not
            a guarantee!.
        :param max_tries: maximum number of re-tries should be made to try and achieve
            the `min_fidelity_all_at_end` constraint.
        :return: list of entanglement info objects per created pair.
        """

        if basis_local is not None:
            rotations_local = basis_to_rotation(basis_local)

        return self.conn.builder.sdk_create_epr_rsp(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=None,
                sequential=False,
                time_unit=time_unit,
                max_time=max_time,
                random_basis_local=random_basis_local,
                rotations_local=rotations_local,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            )
        )

    def create(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        tp: EPRType = EPRType.K,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
        basis_local: EprMeasBasis = None,
        basis_remote: EprMeasBasis = None,
        rotations_local: Tuple[int, int, int] = (0, 0, 0),
        rotations_remote: Tuple[int, int, int] = (0, 0, 0),
        random_basis_local: Optional[RandomBasis] = None,
        random_basis_remote: Optional[RandomBasis] = None,
    ) -> Union[List[Qubit], List[EprMeasureResult], List[LinkLayerOKTypeM]]:
        """Ask the network stack to generate EPR pairs with the remote node.

        A `create` operation must always be matched by a `recv` operation on the remote
        node.

        If the type of request is Create and Keep (CK, or just K) and if `sequential`
        is False (default), this operation returns a list of Qubit objects representing
        the local qubits that are each one half of the generated pairs. These qubits
        can then be manipulated locally just like locally initialized qubits, by e.g.
        applying gates or measuring them.
        Each qubit also contains information about the entanglement generation that
        lead to its creation, and can be accessed by its `entanglement_info` property.

        A typical example for just generating one pair with another node would be:

        .. code-block::

            q = epr_socket.create()[0]
            # `q` can now be used as a normal qubit

        If the type of request is Measure Directly (MD, or just M), this operation
        returns a list of Linklayer response objects. These objects contain information
        about the entanglement generation and includes the measurement outcome and
        basis used. Note that all values are `Future` objects. This means that the
        current subroutine must be flushed before the values become defined.

        An example for generating 10 pairs with another node that are immediately
        measured:

        .. code-block::

            # list of Futures that become defined when subroutine is flushed
            outcomes = []
            with NetQASMConnection("alice", epr_sockets=[epr_socket]):
                ent_infos = epr_socket.create(number=10, tp=EPRType.M)
                for ent_info in ent_infos:
                    outcomes.append(ent_info.measurement_outcome)

        For "Measure Directly"-type requests, the basis to measure in can also be
        specified. There are 3 ways to specify a basis:

        * using one of the `EprMeasBasis` variants
        * by specifying 3 rotation angles, interpreted as an X-rotation, a Y-rotation
          and another X-rotation. For example, setting `rotations_local` to (8, 0, 0)
          means that before measuring, an X-rotation of 8*pi/16 = pi/2 radians is
          applied to the qubit.
        * using one of the `RandomBasis` variants, in which case one of the bases of
          that variant is chosen at random just before measuring

        NOTE: the node that initiates the entanglement generation, i.e. the one that
        calls `create` on its EPR socket, also controls the measurement bases of the
        receiving node (by setting e.g. `rotations_remote`). The receiving node cannot
        change this.

        If `sequential` is False (default), the all requested EPR pairs are generated
        at once, before returning the results (qubits or entanglement info objects).

        If `sequential` is True, a callback function (`post_routine`) should be
        specified. After generating one EPR pair, this callback will be called, before
        generating the next pair. This method can e.g. be used to generate many EPR
        pairs (more than the number of physical qubits available), by measuring (and
        freeing up) each qubit before the next pair is generated.

        For example:

        .. code-block::

            outcomes = alice.new_array(num)

            def post_create(conn, q, pair):
                q.H()
                outcome = outcomes.get_future_index(pair)
                q.measure(outcome)
            epr_socket.create(number=num, post_routine=post_create, sequential=True)


        :param number: number of EPR pairs to generate, defaults to 1
        :param post_routine: callback function for each genated pair. Only used if
            `sequential` is True.
            The callback should take three arguments `(conn, q, pair)` where
            * `conn` is the connection (e.g. `self`)
            * `q` is the entangled qubit (of type `FutureQubit`)
            * `pair` is a register holding which pair is handled (0, 1, ...)
        :param sequential: whether to use callbacks after each pair, defaults to False
        :param tp: type of entanglement generation, defaults to EPRType.K. Note that
            corresponding `recv` of the remote node's EPR socket must specify the
            same type.
        :param time_unit: which time unit to use for the `max_time` parameter
        :param max_time: maximum number of time units (see `time_unit`) the Host is
            willing to wait for entanglement generation of a single pair. If generation
            does not succeed within this time, the whole subroutine that this request
            is part of is reset and run again by the quantum node controller.
        :param basis_local: basis to measure in on this node for M-type requests
        :param basis_remote: basis to measure in on the remote node for M-type requests
        :param rotations_local: rotations to apply before measuring on this node
            (for M-type requests)
        :param rotations_remote: rotations to apply before measuring on remote node
            (for M-type requests)
        :param random_basis_local: random bases to choose from when measuring on this
            node (for M-type requests)
        :param random_basis_remote: random bases to choose from when measuring on
            the remote node (for M-type requests)
        :return: For K-type requests: list of qubits created. For M-type requests:
            list of entanglement info objects per created pair.
        """

        self._logger.warning(
            "EPRSocket.create() is deprecated. Use one of "
            "create_keep, create_measure, or create_rsp instead."
        )

        if tp == EPRType.K:
            return self.create_keep(
                number=number,
                post_routine=post_routine,
                sequential=sequential,
                time_unit=time_unit,
                max_time=max_time,
            )
        elif tp == EPRType.M:
            return self.create_measure(
                number=number,
                time_unit=time_unit,
                max_time=max_time,
                basis_local=basis_local,
                basis_remote=basis_remote,
                rotations_local=rotations_local,
                rotations_remote=rotations_remote,
                random_basis_local=random_basis_local,
                random_basis_remote=random_basis_remote,
            )
        elif tp == EPRType.R:
            return self.create_rsp(
                number=number,
                time_unit=time_unit,
                max_time=max_time,
                basis_local=basis_local,
                random_basis_local=random_basis_local,
            )
        assert False

    def create_context(
        self,
        number: int = 1,
        sequential: bool = False,
        time_unit: TimeUnit = TimeUnit.MICRO_SECONDS,
        max_time: int = 0,
    ) -> ContextManager[Tuple[FutureQubit, RegFuture]]:
        """Create a context that is executed for each generated EPR pair consecutively.

        Creates EPR pairs with a remote node and handles each pair by
        the operations defined in a subsequent context. See the example below.

        .. code-block::

            with epr_socket.create_context(number=10) as (q, pair):
                q.H()
                m = q.measure()

        NOTE: even though all pairs are handled consecutively, they are still
        generated concurrently by the network stack. By setting `sequential` to True,
        the network stack only generates the next pair after the context for the
        previous pair has been executed, similar to using a callback (`post_routine`)
        in the `create` method.

        :param number: number of EPR pairs to generate, defaults to 1
        :param sequential: whether to generate pairs sequentially, defaults to False
        """
        return self.conn.builder.sdk_create_epr_context(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=None,
                sequential=sequential,
                time_unit=time_unit,
                max_time=max_time,
            )
        )

    def recv_keep(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        expect_phi_plus: bool = True,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> List[Qubit]:
        """Ask the network stack to wait for the remote node to generate EPR pairs,
        which are kept in memory.

        A `recv_keep` operation must always be matched by a `create_keep` operation on
        the remote node. The number of generated pairs must also match.

        For more information see the documentation of `create_keep`.

        :param number: number of pairs to generate, defaults to 1
        :param post_routine: callback function used when `sequential` is True
        :param sequential: whether to call the callback after each pair generation,
            defaults to False
        :param expect_phi_plus: whether to assume that the EPR pairs that are created
            are in the Phi+ (or Phi_00) state. Defaults to True. If True, the compiler
            will make sure that if the physical link actually produced another Bell
            state, the behavior seen by the application is still as if a Phi+ state
            was actually produced.
        :param min_fidelity_all_at_end: the minimum fidelity that *all* entangled
            qubits should ideally still have at the moment the last qubit has been
            generated. For example, when specifying `number=2` and
            `min_fidelity_all_at_end=80`, the the program will automatically try to
            make sure that both qubits have a fidelity of at least 80% when the
            second qubit has been generated. It will attempt to do this by
            automatically re-trying the entanglement generation if the fidelity
            constraint is not satisfied. This is however an *attempt*, and not
            a guarantee!.
        :param max_tries: maximum number of re-tries should be made to try and achieve
            the `min_fidelity_all_at_end` constraint.
        :return: list of qubits created
        """

        if self.conn is None:
            raise RuntimeError("EPRSocket does not have an open connection")

        qubits, _ = self.conn._builder.sdk_recv_epr_keep(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=post_routine,
                sequential=sequential,
                expect_phi_plus=expect_phi_plus,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            ),
        )
        return qubits

    def recv_keep_with_info(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        expect_phi_plus: bool = True,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Same as recv_keep but also return the EPR generation information coming
        from the network stack.

        For more information see the documentation of `recv_keep`.

        :param number: number of pairs to generate, defaults to 1
        :return: tuple with (1) list of qubits created, (2) list of EprKeepResult objects
        """
        qubits, info = self.conn._builder.sdk_recv_epr_keep(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=post_routine,
                sequential=sequential,
                expect_phi_plus=expect_phi_plus,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            ),
        )
        return qubits, info

    def recv_measure(
        self,
        number: int = 1,
        expect_phi_plus: bool = True,
    ) -> List[EprMeasureResult]:
        """Ask the network stack to wait for the remote node to generate EPR pairs,
        which are immediately measured (on both nodes).

        A `recv_measure` operation must always be matched by a `create_measure`
        operation on the remote node. The number and type of generation must also match.

        For more information see the documentation of `create_measure`.

        :param number: number of pairs to generate, defaults to 1
        :param expect_phi_plus: whether to assume that the EPR pairs that are created
            are in the Phi+ (or Phi_00) state. Defaults to True. If True, the compiler
            will make sure that if the physical link actually produced another Bell
            state, the behavior seen by the application is still as if a Phi+ state
            was actually produced.
        :return: list of entanglement info objects per created pair.
        """
        if self.conn is None:
            raise RuntimeError("EPRSocket does not have an open connection")

        return self.conn.builder.sdk_recv_epr_measure(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                expect_phi_plus=expect_phi_plus,
                post_routine=None,
                sequential=False,
            ),
        )

    def recv_rsp(
        self,
        number: int = 1,
        expect_phi_plus: bool = True,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> List[Qubit]:
        """Ask the network stack to wait for remote state preparation from another node.

        A `recv_rsp` operation must always be matched by a `create_rsp` operation on
        the remote node. The number and type of generation must also match.

        For more information see the documentation of `create_rsp`.

        :param number: number of pairs to generate, defaults to 1
        :param expect_phi_plus: whether to assume that the EPR pairs that are created
            are in the Phi+ (or Phi_00) state. Defaults to True. If True, the compiler
            will make sure that if the physical link actually produced another Bell
            state, the behavior seen by the application is still as if a Phi+ state
            was actually produced.
        :param min_fidelity_all_at_end: the minimum fidelity that *all* entangled
            qubits should ideally still have at the moment the last qubit has been
            generated. For example, when specifying `number=2` and
            `min_fidelity_all_at_end=80`, the the program will automatically try to
            make sure that both qubits have a fidelity of at least 80% when the
            second qubit has been generated. It will attempt to do this by
            automatically re-trying the entanglement generation if the fidelity
            constraint is not satisfied. This is however an *attempt*, and not
            a guarantee!.
        :param max_tries: maximum number of re-tries should be made to try and achieve
            the `min_fidelity_all_at_end` constraint.
        :return: list of qubits created
        """
        if self.conn is None:
            raise RuntimeError("EPRSocket does not have an open connection")

        qubits, _ = self.conn.builder.sdk_recv_epr_rsp(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=None,
                sequential=False,
                expect_phi_plus=expect_phi_plus,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            ),
        )
        return qubits

    def recv_rsp_with_info(
        self,
        number: int = 1,
        expect_phi_plus: bool = True,
        min_fidelity_all_at_end: Optional[int] = None,
        max_tries: Optional[int] = None,
    ) -> Tuple[List[Qubit], List[EprKeepResult]]:
        """Same as recv_rsp but also return the EPR generation information coming
        from the network stack.

        For more information see the documentation of `recv_rsp`.

        :param number: number of pairs to generate, defaults to 1
        :return: tuple with (1) list of qubits created, (2) list of EprKeepResult objects
        """
        if self.conn is None:
            raise RuntimeError("EPRSocket does not have an open connection")

        qubits, infos = self.conn.builder.sdk_recv_epr_rsp(
            params=EntRequestParams(
                remote_node_id=self.remote_node_id,
                epr_socket_id=self._epr_socket_id,
                number=number,
                post_routine=None,
                sequential=False,
                expect_phi_plus=expect_phi_plus,
                min_fidelity_all_at_end=min_fidelity_all_at_end,
                max_tries=max_tries,
            ),
        )
        return qubits, infos

    def recv(
        self,
        number: int = 1,
        post_routine: Optional[Callable] = None,
        sequential: bool = False,
        tp: EPRType = EPRType.K,
    ) -> Union[List[Qubit], List[EprMeasureResult], List[LinkLayerOKTypeR]]:
        """Ask the network stack to wait for the remote node to generate EPR pairs.

        A `recv` operation must always be matched by a `create` operation on the remote
        node. See also the documentation of `create`.
        The number and type of generation must also match.

        In case of Measure Directly requests, it is the initiating node (that calls
        `create`) which specifies the measurement bases. This should not and cannot be
        done in `recv`.

        For more information see the documentation of `create`.

        :param number: number of pairs to generate, defaults to 1
        :param post_routine: callback function used when `sequential` is True
        :param sequential: whether to call the callback after each pair generation,
            defaults to False
        :param tp: type of entanglement generation, defaults to EPRType.K
        :return: For K-type requests: list of qubits created. For M-type requests:
            list of entanglement info objects per created pair.
        """
        self._logger.warning(
            "EPRSocket.recv() is deprecated. Use one of "
            "recv_keep, recv_measure, or recv_rsp instead."
        )

        if tp == EPRType.K:
            return self.recv_keep(
                number=number,
                post_routine=post_routine,
                sequential=sequential,
            )
        elif tp == EPRType.M:
            return self.recv_measure(number=number)
        elif tp == EPRType.R:
            return self.recv_rsp(number=number)
        assert False

    @contextmanager
    def recv_context(
        self,
        number: int = 1,
        sequential: bool = False,
    ):
        """Receives EPR pair with a remote node (see doc of :meth:`~.create_context`)"""
        try:
            # NOTE loop_register is the register used for looping over the generated pairs
            (
                pre_commands,
                loop_register,
                ent_results_array,
                output,
                pair,
            ) = self.conn.builder._pre_epr_context(
                role=EPRRole.RECV,
                params=EntRequestParams(
                    remote_node_id=self.remote_node_id,
                    epr_socket_id=self._epr_socket_id,
                    number=number,
                    post_routine=None,
                    sequential=sequential,
                ),
            )
            yield output, pair
        finally:
            self.conn.builder._post_epr_context(
                pre_commands=pre_commands,
                number=number,
                loop_register=loop_register,
                ent_results_array=ent_results_array,
                pair=pair,
            )

    def _get_node_id(self, app_name: str) -> int:
        return self.conn.network_info.get_node_id_for_app(app_name=app_name)



from __future__ import annotations

from itertools import count
from typing import Dict, List, Set

from netqasm.lang import operand
from netqasm.lang.encoding import REG_INDEX_BITS, RegisterName
from netqasm.lang.parsing.text import parse_register
from netqasm.sdk.futures import Array
from netqasm.sdk.qubit import Qubit


class MemoryManager:
    """Container for managing application memory during building.

    Used by a Builder to store information about which registers, arrays,
    and qubits have been allocated, and to (de)allocate these memory locations
    when needed.
    """

    def __init__(self) -> None:
        # All qubits active for this connection
        self._active_qubits: List[Qubit] = []

        # Registers that are in use for holding classical data.
        self._active_registers: Set[operand.Register] = set()

        # Register in use for holding measurement outcomes.
        self._used_meas_registers: Dict[operand.Register, bool] = {
            operand.Register(RegisterName.M, i): False for i in range(16)
        }

        # Registers that need to be returned at the end of the subroutine.
        self._registers_to_return: List[operand.Register] = []

        # Addresses that are already used for allocated arrays.
        self._used_array_addresses: List[int] = []

        # Arrays that need to be returned at the end of the subroutine.
        self._arrays_to_return: List[Array] = []

    def inactivate_qubits(self) -> None:
        """Mark all registers as inactive (i.e. not in use)."""
        while len(self._active_qubits) > 0:
            q = self._active_qubits.pop()
            q.active = False

    def get_active_qubits(self) -> List[Qubit]:
        """Get all qubit locations that are in use."""
        return self._active_qubits

    def is_qubit_active(self, q: Qubit) -> bool:
        """Check if a qubit location is in use."""
        return q in self._active_qubits

    def is_qubit_id_used(self, id: int) -> bool:
        """Check if a qubit ID is in use."""
        return any(q.qubit_id == id for q in self._active_qubits)

    def activate_qubit(self, q: Qubit) -> None:
        """Mark a qubit location as 'in use'."""
        self._active_qubits.append(q)

    def deactivate_qubit(self, q: Qubit) -> None:
        """Mark a qubit location as 'not in use'."""
        self._active_qubits.remove(q)

    def get_new_qubit_address(self) -> int:
        """Get an unused qubit location."""
        qubit_addresses_in_use = [q.qubit_id for q in self._active_qubits]
        for address in count(0):
            if address not in qubit_addresses_in_use:
                return address
        raise RuntimeError("Could not get new qubit address")

    def is_register_active(self, reg: operand.Register) -> bool:
        """Check if a register is in use."""
        return reg in self._active_registers

    def add_active_register(self, reg: operand.Register) -> None:
        """Mark a register as 'in use'."""
        if reg in self._active_registers:
            raise ValueError(f"Register {reg} is already active")
        self._active_registers.add(reg)

    def remove_active_register(self, reg: operand.Register) -> None:
        """Mark a register as 'not in use'."""
        self._active_registers.remove(reg)

    def meas_register_set_used(self, reg: operand.Register) -> None:
        """Mark a measurement register as 'in use'."""
        self._used_meas_registers[reg] = True

    def meas_register_set_unused(self, reg: operand.Register) -> None:
        """Mark a measurement register as 'not in use'."""
        self._used_meas_registers[reg] = False

    def get_new_meas_outcome_register(self) -> operand.Register:
        """Get an un-used measurement register."""
        # Find the next unused M-register.
        for reg, used in self._used_meas_registers.items():
            if not used:
                self._used_meas_registers[reg] = True
                return reg
        raise RuntimeError("Ran out of M-registers")

    def reset_used_meas_registers(self) -> None:
        """Mark all measurement registers as 'not in use'."""
        self._used_meas_registers = {
            operand.Register(RegisterName.M, i): False for i in range(16)
        }

    def add_register_to_return(self, reg: operand.Register) -> None:
        """Let a register be returned at the end of the subroutine."""
        self._registers_to_return.append(reg)

    def get_registers_to_return(self) -> List[operand.Register]:
        """Get all register that are returned at the end of the subroutine."""
        return self._registers_to_return

    def reset_registers_to_return(self) -> None:
        """Clear list of registers that are returned at the end of the subroutine."""
        self._registers_to_return = []

    def get_inactive_register(self, activate: bool = False) -> operand.Register:
        """Get an un-used register."""
        for i in range(2**REG_INDEX_BITS):
            register = parse_register(f"R{i}")
            if not self.is_register_active(register):
                if activate:
                    self.add_active_register(register)
                return register
        raise RuntimeError("could not find an available loop register")

    def get_new_array_address(self) -> int:
        """Get an un-used array address."""
        if len(self._used_array_addresses) > 0:
            # last element is always the highest address
            address = self._used_array_addresses[-1] + 1
        else:
            address = 0
        self._used_array_addresses.append(address)
        return address

    def add_array_to_return(self, array: Array) -> None:
        """Let an array be returned at the end of the subroutine."""
        self._arrays_to_return.append(array)

    def get_arrays_to_return(self) -> List[Array]:
        """Get all arrays that are returned at the end of the subroutine."""
        return self._arrays_to_return

    def reset_arrays_to_return(self) -> None:
        """Clear list of arrays that are returned at the end of the subroutine."""
        self._arrays_to_return = []

    def reset(self) -> None:
        """Reset the state of the MemoryManager."""
        self.reset_arrays_to_return()
        self.reset_registers_to_return()
        self.reset_used_meas_registers()



import random

from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    with NetQASMConnection("Alice") as alice:
        num = 10

        outcomes = alice.new_array(num)
        rand_nums = alice.new_array(
            num, init_values=[random.randint(0, 1) for _ in range(num)]
        )
        i = alice.new_array(1, init_values=[0]).get_future_index(0)

        with rand_nums.foreach() as r:
            q = Qubit(alice)
            with r.if_eq(1):
                q.X()
            q.measure(future=outcomes.get_future_index(i))
            i.add(1)

        alice.flush()
        assert len(outcomes) == num
        print(f"rand_nums = {list(rand_nums)}")
        print(f"outcomes = {list(outcomes)}")
        assert list(rand_nums) == list(outcomes)


def test_foreach():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



from netqasm.examples.lib.bqc import (
    measXY,
    recv_meas_cmd,
    recv_teleported_state,
    send_meas_outcome,
)
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state

logger = get_netqasm_logger()


def main(app_config=None):
    socket = Socket("server", "client", log_config=app_config.log_config)
    epr_socket = EPRSocket("client")

    server = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    with server:
        # Let client remotely prepare two qubits.
        q1 = recv_teleported_state(epr_socket)
        q2 = recv_teleported_state(epr_socket)

        server.flush()

        socket.recv_silent()

        # Apply a CPHASE gate between the two qubits.
        q1.cphase(q2)

        # TODO check why this is needed
        server.flush()

        # Receive from the client the angle to measure the first qubit in.
        angle = recv_meas_cmd(socket)
        s = measXY(q1, angle)
        server.flush()
        send_meas_outcome(socket, s)

        # Receive from the client the angle to measure the first qubit in.
        angle = recv_meas_cmd(socket)
        server.flush()
        q2.rot_Z(angle=angle)
        q2.H()
        server.flush()
        dm = get_qubit_state(q2)
        s = q2.measure()
        server.flush()
        send_meas_outcome(socket, s)

        return {"final_state": dm if dm is None else dm.tolist()}



"""
Interface to quantum node controllers.

This module provides the `BaseNetQASMConnection` class which represents
the connection with a quantum node controller.
"""

from __future__ import annotations

import abc
import logging
import math
import os
import pickle
from itertools import count
from typing import (
    TYPE_CHECKING,
    Callable,
    ContextManager,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from netqasm.backend.messages import (
    InitNewAppMessage,
    Message,
    OpenEPRSocketMessage,
    Signal,
    SignalMessage,
    StopAppMessage,
    SubroutineMessage,
)
from netqasm.lang import operand
from netqasm.lang.ir import BreakpointAction, BreakpointRole, ProtoSubroutine
from netqasm.lang.subroutine import Subroutine
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk.build_types import (
    GenericHardwareConfig,
    HardwareConfig,
    T_BranchRoutine,
    T_LoopRoutine,
)
from netqasm.sdk.config import LogConfig
from netqasm.sdk.futures import Array, Future, RegFuture, T_CValue
from netqasm.sdk.network import NetworkInfo
from netqasm.sdk.progress_bar import ProgressBar
from netqasm.sdk.qubit import Qubit
from netqasm.sdk.shared_memory import SharedMemory, SharedMemoryManager
from netqasm.sdk.transpile import SubroutineTranspiler
from netqasm.util.log import LineTracker

from .builder import Builder, SdkLoopUntilContext

# Generic type for messages sent to the quantum node controller.
# Note that `SubroutineMessage` does not derive from `Message` so it has to be
# mentioned explicitly.
T_Message = Union[Message, SubroutineMessage]

# Imports that are only needed for type checking
if TYPE_CHECKING:
    from netqasm.sdk import epr_socket as esck


class BaseNetQASMConnection(abc.ABC):
    """Base class for representing connections to a quantum node controller.

    A BaseNetQASMConnection instance provides an interface for Host programs to
    interact with a quantum node controller like QNodeOS, which controls the quantum
    hardware.

    The interaction with the quantum node controller includes registering applications,
    opening EPR sockets, sending NetQASM subroutines, and getting execution results,

    A BaseNetQASMConnection instance also provides a 'context' for the Host to
    run its code in. Code within this context is compiled into NetQASM subroutines
    and then sent to the quantum node controller.
    """

    # Global dict to track all used app IDs for each program
    _app_ids: Dict[str, List[int]] = {}  # <party> -> [<app_id1>, <app_id2>, ...]

    # Global dict to track app names per node.
    # Currently only used for logging purposes, specifically only in
    # `netqasm.runtime.process_logs._create_app_instr_logs`.
    # Dict[node_name, Dict[app_id, app_name]]
    _app_names: Dict[str, Dict[int, str]] = {}

    def __init__(
        self,
        app_name: str,
        node_name: Optional[str] = None,
        app_id: Optional[int] = None,
        max_qubits: int = 5,
        hardware_config: Optional[HardwareConfig] = None,
        log_config: LogConfig = None,
        epr_sockets: Optional[List[esck.EPRSocket]] = None,
        compiler: Optional[Type[SubroutineTranspiler]] = None,
        return_arrays: bool = True,
        _init_app: bool = True,
        _setup_epr_sockets: bool = True,
    ):
        """BaseNetQASMConnection constructor.

        In most cases, you will want to instantiate a subclass of this.

        :param app_name: Name of the application.
            Specifically, this is the name of the program that runs on this particular
            node. So, `app_name` can often be seen as the name of the "role" within the
            multi-party application or protocol.
            For example, in a Blind Computation protocol, the two roles may be
            "client" and "server"; the `app_name` of a particular
            `BaseNetQASMConnection` instance may then e.g. be "client".

        :param node_name: name of the Node that is controlled by the quantum
            node controller that we connect to. The Node name may be different from
            the `app_name`, and e.g. reflect its geographical location, like a city
            name. If None, the Node name is obtained by querying the global
            `NetworkInfo` by using the `app_name`.

        :param app_id: ID of this application.
            An application registered in the quantum node controller using this
            connection will use this app ID.
            If None, a unique ID will be chosen (unique among possible other
            applications that were registered through other BaseNetQASMConnection
            instances).

        :param max_qubits: maximum number of qubits that can be in use
            at the same time by applications registered through this connection.
            Defaults to 5.

        :param hardware_config: configuration object with info about the underlying
            hardware. Used by the Builder of this Connection. When None,
            a generic configuration object is created with the default qubit count of 5.

        :param log_config: configuration object specifying what to log.

        :param epr_sockets: list of EPR sockets.
            If `_setup_epr_sockets` is True, these EPR sockets are automatically
            opened upon entering this connection's context.

        :param compiler: the class that is used to instantiate the compiler.
            If None, a compiler is used that can compile for the hardware of the
            node this connection is to.

        :param return_arrays: whether to add "return array"-instructions at the end
            of subroutines. A reason to set this to False could be that a quantum
            node controller does not support returning arrays back to the Host.

        :param _init_app: whether to immediately send a "register application" message
            to the quantum node controller upon construction of this connection.

        :param _setup_epr_sockets: whether to immediately send "open EPR socket"
            messages to the quantum node controller upon construction of this
            connection. If True, the "open EPR socket" messages are for the EPR sockets
            defined in the `epr_sockets` parameter.
        """
        self._app_name: str = app_name

        # Set the app ID. If one was provided in `app_id`, try to use that one.
        self._app_id: int = self._get_new_app_id(app_id)

        if node_name is None:
            node_name = self.network_info.get_node_name_for_app(app_name)
        self._node_name: str = node_name

        # Update global _app_names dict. Only used for logging purposes.
        if node_name not in self._app_names:
            self._app_names[node_name] = {}
        self._app_names[node_name][self._app_id] = app_name

        # Try to obtain the Shared Memory instance corresponding to this connection.
        # Depending on the runtime environment, this instance may not yet exist at
        # this moment. If this is the case, the SharedMemory instance *should*
        # become available later on.
        self._shared_memory: Optional[
            SharedMemory
        ] = SharedMemoryManager.get_shared_memory(self.node_name, key=self._app_id)

        # Whether to send additional closing messages to the quantum node controller
        # at the end of an application. See the `close()` method.
        # At the moment only used for debugging and not exposed to the user.
        self._clear_app_on_exit: bool = True
        self._stop_backend_on_exit: bool = True

        if log_config is None:
            log_config = LogConfig()

        self._line_tracker: LineTracker = LineTracker(log_config=log_config)
        self._track_lines: bool = log_config.track_lines

        # Set directory for storing serialized subroutines.
        self._log_subroutines_dir: Optional[str] = log_config.log_subroutines_dir

        if hardware_config is None:
            hardware_config = GenericHardwareConfig(max_qubits)

        # Builder object of this connection.
        # The Builder is used to gather application code and output an
        # Intermediate Representation (IR), which is then compiled into subroutines.
        self._builder = Builder(
            connection=self,
            app_id=self._app_id,
            hardware_config=hardware_config,
            compiler=compiler,
            return_arrays=return_arrays,
        )

        # What compiler (if any) to be used.
        # Currently, the "_compiler" attribute is more of a transpiler, since it can
        # only convert between NetQASM flavours.
        # The actual conversion from Python DSL statements to IR to NetQASM is
        # done by the Builder.
        self._compiler: Optional[Type[SubroutineTranspiler]] = compiler

        # Logger for stdout. (Not for log files.)
        self._logger: logging.Logger = get_netqasm_logger(
            f"{self.__class__.__name__}({self.app_name})"
        )

        if _init_app:
            self._init_new_app(max_qubits=max_qubits)

        if _setup_epr_sockets:
            self._setup_epr_sockets(epr_sockets=epr_sockets)

    @property
    def app_name(self) -> str:
        """Get the application name"""
        return self._app_name

    @property
    def node_name(self) -> str:
        """Get the node name"""
        return self._node_name

    @property
    def app_id(self) -> int:
        """Get the application ID"""
        return self._app_id

    @app_id.setter
    def app_id(self, id: int) -> None:
        self._app_id = id
        self._builder.app_id = id

    @abc.abstractmethod
    def _get_network_info(self) -> Type[NetworkInfo]:
        raise NotImplementedError

    @property
    def network_info(self) -> Type[NetworkInfo]:
        return self._get_network_info()

    @property
    def builder(self) -> Builder:
        return self._builder

    @classmethod
    def get_app_ids(cls) -> Dict[str, List[int]]:
        return cls._app_ids

    @classmethod
    def get_app_names(cls) -> Dict[str, Dict[int, str]]:
        return cls._app_names

    def __str__(self):
        return (
            f"NetQASM connection for app '{self.app_name}' with node '{self.node_name}'"
        )

    def __enter__(self):
        """Start a context with this connection.

        Quantum operations specified within the connection are automatically compiled
        into NetQASM subroutines.
        These subroutines are sent to the quantum node controller, over this
        connection, when either :meth:`~.flush` is called or the connection goes out
        of context which calls :meth:`~.__exit__`.

        .. code-block::

            # Open the connection
            with NetQASMConnection(app_name="alice") as alice:
                # Create a qubit
                q = Qubit(alice)
                # Perform a Hadamard
                q.H()
                # Measure the qubit
                m = q.measure()
                # Flush the subroutine to populate the variable `m` with the outcome
                # Alternatively, this can be done by letting the connection
                # go out of context and move the print to after.
                alice.flush()
                print(m)
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Called automatically when a connection context ends.

        Default behavior is to call the `close` method on the connection.
        """
        # Allow to not clear the app or stop the backend upon exit, for debugging and post processing
        self.close(
            clear_app=self._clear_app_on_exit,
            stop_backend=self._stop_backend_on_exit,
        )

    def _get_new_app_id(self, app_id: Optional[int]) -> int:
        """Find a suitable new app ID.

        Tries to use the specified `app_id`, if not in use already.
        """
        name = self.app_name
        if name not in self._app_ids:
            self._app_ids[name] = []

        # Which app_id
        if app_id is None:
            for app_id in count(0):
                if app_id not in self._app_ids[name]:
                    self._app_ids[name].append(app_id)
                    return app_id
            raise RuntimeError("This should never be reached")
        else:
            if app_id in self._app_ids[name]:
                raise ValueError("app_id={} is already in use".format(app_id))
            self._app_ids[name].append(app_id)
            return app_id

    def _pop_app_id(self) -> None:
        """Remove the used app ID from the list."""
        try:
            self._app_ids[self.app_name].remove(self.app_id)
        except ValueError:
            pass  # Already removed

    def clear(self) -> None:
        self._pop_app_id()

    def close(self, clear_app: bool = True, stop_backend: bool = False) -> None:
        """Close a connection.

        By default, this method is automatically called when a connection context ends.
        """
        # Flush all pending commands
        self.flush()

        self._pop_app_id()

        self._signal_stop(clear_app=clear_app, stop_backend=stop_backend)
        self._builder.inactivate_qubits()

        if self._log_subroutines_dir is not None:
            self._save_log_subroutines()

    def _commit_message(
        self, msg: T_Message, block: bool = True, callback: Optional[Callable] = None
    ) -> None:
        """Commit a message to the quantum node controller.

        The message gets serialized and then sent through the connection.
        """
        self._logger.debug(f"Committing message {msg}")
        self._commit_serialized_message(
            raw_msg=bytes(msg), block=block, callback=callback
        )

    @abc.abstractmethod
    def _commit_serialized_message(
        self, raw_msg: bytes, block: bool = True, callback: Optional[Callable] = None
    ) -> None:
        """Commit a serialized message to the quantum node controller."""
        pass

    def _signal_stop(self, clear_app: bool = True, stop_backend: bool = True) -> None:
        """Signal to the quantum node controller to stop.

        This signal may be used with simulator backends so that they can e.g. reset
        their state.
        """
        if clear_app:
            self._commit_message(msg=StopAppMessage(app_id=self._app_id))

        if stop_backend:
            self._commit_message(msg=SignalMessage(signal=Signal.STOP), block=False)

    def _save_log_subroutines(self) -> None:
        """Serialize and dump subroutines that were sent over this connection."""
        filename = f"subroutines_{self.app_name}.pkl"
        filepath = os.path.join(self._log_subroutines_dir, filename)  # type: ignore
        with open(filepath, "wb") as f:
            pickle.dump(self._builder.committed_subroutines, f)

    @property
    def shared_memory(self) -> SharedMemory:
        """Get this connection's Shared Memory object.

        This property should *not* be accessed before any potential setting-up
        of shared memories has finished. If it cannot be found, an error is raised.
        """
        if self._shared_memory is None:
            mem = SharedMemoryManager.get_shared_memory(
                self.node_name, key=self._app_id
            )
            if mem is None:
                raise RuntimeError(
                    "Trying to access connection's shared memory, but it does not exist"
                )
            self._shared_memory = mem
        return self._shared_memory

    @property
    def active_qubits(self) -> List[Qubit]:
        """Get a list of qubits that are currently in use.

        "In use" means that the virtual qubit represented by this `Qubit` instance
        has been allocated and hence its virtual ID cannot be re-used.

        :return: list of active qubits
        """
        return self._builder._mem_mgr.get_active_qubits()

    def _init_new_app(self, max_qubits: int) -> None:
        """Send a message to the quantum node controller to register a new application.

        :param max_qubits: maximum number of qubits that this application may allocate
            at the same time.
        """
        self._commit_message(
            msg=InitNewAppMessage(
                app_id=self._app_id,
                max_qubits=max_qubits,
            )
        )

        # Wait until a shared memory is available.
        # This is to be sure that we can access the shared memory at any time
        # during the application.
        # TODO: improve how this is handled
        while self.shared_memory is None:
            pass

    def _setup_epr_sockets(self, epr_sockets: Optional[List[esck.EPRSocket]]) -> None:
        """Send messages to the quantum node controller to open EPR sockets."""
        if epr_sockets is None:
            return
        for epr_socket in epr_sockets:
            if epr_socket._remote_app_name == self.app_name:
                raise ValueError("A node cannot setup an EPR socket with itself")
            epr_socket.conn = self
            self._setup_epr_socket(
                epr_socket_id=epr_socket.epr_socket_id,
                remote_node_id=epr_socket.remote_node_id,
                remote_epr_socket_id=epr_socket.remote_epr_socket_id,
                min_fidelity=epr_socket.min_fidelity,
            )

    def _setup_epr_socket(
        self,
        epr_socket_id: int,
        remote_node_id: int,
        remote_epr_socket_id: int,
        min_fidelity: int,
    ) -> None:
        """Send a message to the quantum node controller to open an EPR socket."""
        self._commit_message(
            msg=OpenEPRSocketMessage(
                app_id=self._app_id,
                epr_socket_id=epr_socket_id,
                remote_node_id=remote_node_id,
                remote_epr_socket_id=remote_epr_socket_id,
                min_fidelity=min_fidelity,
            )
        )

    def flush(self, block: bool = True, callback: Optional[Callable] = None) -> None:
        """Compile and send all pending operations to the quantum node controller.

        All operations that have been added to this connection's Builder (typically by
        issuing these operations within a connection context) are collected and
        compiled into a NetQASM subroutine. This subroutine is then sent over the connection
        to be executed by the quantum node controller.

        :param block: block on receiving the result of executing the compiled subroutine
            from the quantum node controller.
        :param callback: if `block` is False, this callback is called when the quantum
            node controller sends the subroutine results.
        """
        protosubroutine = self._builder.subrt_pop_pending_subroutine()
        if protosubroutine is None:
            return

        self.commit_protosubroutine(
            protosubroutine=protosubroutine,
            block=block,
            callback=callback,
        )

    def compile(self) -> Optional[Subroutine]:
        """Compile the previous SDK commands into a NetQASM subroutine.

        This does this the same as calling `flush()`, except it does not send the
        subroutine to the quantum node controller for execution. This method can hence
        be used to pre-compile a subroutine and send it later, possibly after filling
        in concrete values for templates.
        """
        protosubroutine = self._builder.subrt_pop_pending_subroutine()
        self._logger.info(f"Compiling protosubroutine:\n{protosubroutine}")
        if protosubroutine is None:
            return None

        subroutine = self._builder.subrt_compile_subroutine(protosubroutine)
        self._builder._reset()

        return subroutine

    def commit_protosubroutine(
        self,
        protosubroutine: ProtoSubroutine,
        block: bool = True,
        callback: Optional[Callable] = None,
    ) -> None:
        """Send a protosubroutine to the quantum node controller.

        Takes a `ProtoSubroutine`, i.e. an intermediate representation of the subroutine
        that comes from the Builder.
        The ProtoSubroutine is compiled into a `Subroutine` instance.
        """
        self._logger.debug(f"Flushing protosubroutine:\n{protosubroutine}")

        # Parse, assembly and possibly compile the subroutine
        subroutine = self._builder.subrt_compile_subroutine(protosubroutine)
        self._logger.info(f"Flushing compiled subroutine:\n{subroutine}")

        subroutine.instantiate(self.app_id)

        # Commit the subroutine to the quantum device
        self.commit_subroutine(subroutine, block, callback)

        self._builder._reset()

    def commit_subroutine(
        self,
        subroutine: Subroutine,
        block: bool = True,
        callback: Optional[Callable] = None,
    ) -> None:
        self._logger.info(f"Committing compiled subroutine:\n{subroutine}")

        self._commit_message(
            msg=SubroutineMessage(subroutine=subroutine),
            block=block,
            callback=callback,
        )

    def block(self) -> None:
        """Block until a flushed subroutines finishes.

        This should be implemented by subclasses.

        :raises NotImplementedError
        """
        raise NotImplementedError

    def new_array(
        self, length: int = 1, init_values: Optional[List[Optional[int]]] = None
    ) -> Array:
        """Allocate a new array in the shared memory.

        This operation is handled by the connection's Builder.
        The Builder make sures the relevant NetQASM instructions end up in the
        subroutine.

        :param length: length of the array, defaults to 1
        :param init_values: list of initial values of the array. If not None, must
            have the same length as `length`.
        :return: a handle to the array that can be used in application code
        """
        return self._builder.alloc_array(length, init_values)

    def loop(
        self,
        stop: int,
        start: int = 0,
        step: int = 1,
        loop_register: Optional[operand.Register] = None,
    ) -> ContextManager[operand.Register]:
        """Create a context for code that gets looped.

        Each iteration of the loop is associated with an index, which starts at 0
        by default. Each iteration the index is increased by `step` (default 1).
        Looping stops when the index reaches `stop`.

        Code inside the context *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        This operation is handled by the connection's Builder.
        The Builder make sures the NetQASM subroutine contains a loop around the
        (compiled) code that is inside the context.

        Example:

        .. code-block::

            with NetQASMConnection(app_name="alice") as alice:
                outcomes = alice.new_array(10)
                with alice.loop(10) as i:
                    q = Qubit(alice)
                    q.H()
                    outcome = outcomes.get_future_index(i)
                    q.measure(outcome)

        :param stop: end of iteration range (exluding)
        :param start: start of iteration range (including), defaults to 0
        :param step: step size of iteration range, defaults to 1
        :param loop_register: specific register to be used for holding the loop index.
            In most cases there is no need to explicitly specify this.
        :return: the context object (to be used in a `with ...` expression)
        """
        return self._builder.sdk_loop_context(stop, start, step, loop_register)

    def loop_body(
        self,
        body: T_LoopRoutine,
        stop: int,
        start: int = 0,
        step: int = 1,
        loop_register: Optional[Union[operand.Register, str]] = None,
    ) -> None:
        """Loop code that is defined in a Python function (body).

        The function to loop should have a single argument with that has the
        `BaseNetQASMConnection` type.

        :param body: function to loop
        :param stop: end of iteration range (exluding)
        :param start: start of iteration range (including), defaults to 0
        :param step: step size of iteration range, defaults to 1
        :param loop_register: specific register to be used for holding the loop index.
        """
        self._builder.sdk_loop_body(body, stop, start, step, loop_register)

    def loop_until(self, max_iterations: int) -> ContextManager[SdkLoopUntilContext]:
        """Create a context with code to be looped until the exit condition is met, or
        the maximum number of tries has been reached.

        The code inside the context is automatically looped (re-run).
        At the end of each iteration, the exit_condition of the context object is
        checked. If the condition holds, the loop exits. Otherwise the loop
        does another iteration. If `max_iterations` iterations have been reached,
        the loop exits anyway.

        Make sure you set the loop_condition on the context object, like e.g.

        .. code-block::

        with connection.loop_until(max_iterations=10) as loop:
            q = Qubit(conn)
            m = q.measure()
            constraint = ValueAtMostConstraint(m, 0)
            loop.set_exit_condition(constraint)


        :param max_iterations: the maximum number of times to loop
        """
        return self.builder.sdk_new_loop_until_context(max_iterations)

    def if_eq(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a == b.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param b: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_eq(a, b, body)

    def if_ne(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a != b.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param b: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_ne(a, b, body)

    def if_lt(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a < b.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param b: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_lt(a, b, body)

    def if_ge(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a > b.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param b: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_ge(a, b, body)

    def if_ez(self, a: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a == 0.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_ez(a, body)

    def if_nz(self, a: T_CValue, body: T_BranchRoutine) -> None:
        """Execute a function if a != 0.

        Code inside the callback function *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        :param a: a classical value
        :param body: function to execute if condition is true
        """
        self._builder.sdk_if_nz(a, body)

    def try_until_success(self, max_tries: int = 1) -> ContextManager[None]:
        """TODO docstring"""
        return self._builder.sdk_try_context(max_tries)

    def tomography(
        self,
        preparation: Callable[[BaseNetQASMConnection], Qubit],
        iterations: int,
        progress: bool = True,
    ) -> Dict[str, float]:
        """
        Does a tomography on the output from the preparation specified.
        The frequencies from X, Y and Z measurements are returned as a tuple (f_X,f_Y,f_Z).

        - **Arguments**

            :preparation:     A function that takes a NetQASMConnection as input and prepares a qubit and returns this
            :iterations:     Number of measurements in each basis.
            :progress_bar:     Displays a progress bar
        """
        outcomes: Dict[str, List[Union[Future, RegFuture]]] = {
            "X": [],
            "Y": [],
            "Z": [],
        }
        if progress:
            bar = ProgressBar(3 * iterations)

        # Measure in X
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            q.H()
            m = q.measure()
            outcomes["X"].append(m)

        # Measure in Y
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            q.K()
            m = q.measure()
            outcomes["Y"].append(m)

        # Measure in Z
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            m = q.measure()
            outcomes["Z"].append(m)

        if progress:
            bar.close()
            del bar

        self.flush()

        freqs = {key: sum(value) / iterations for key, value in outcomes.items()}
        return freqs

    def test_preparation(
        self,
        preparation: Callable[[BaseNetQASMConnection], Qubit],
        exp_values: Tuple[float, float, float],
        conf: float = 2,
        iterations: int = 100,
        progress: bool = True,
    ) -> bool:
        """Test the preparation of a qubit.
        Returns True if the expected values are inside the confidence interval produced from the data received from
        the tomography function

        - **Arguments**

            :preparation:     A function that takes a NetQASMConnection as input and prepares a qubit and returns this
            :exp_values:     The expected values for measurements in the X, Y and Z basis.
            :conf:         Determines the confidence region (+/- conf/sqrt(iterations) )
            :iterations:     Number of measurements in each basis.
            :progress_bar:     Displays a progress bar
        """
        epsilon = conf / math.sqrt(iterations)

        freqs = self.tomography(preparation, iterations, progress=progress)
        for basis, exp_value in zip(["X", "Y", "Z"], exp_values):
            f = freqs[basis]
            if abs(f - exp_value) > epsilon:
                print(freqs, exp_values, epsilon)
                return False
        return True

    def insert_breakpoint(
        self, action: BreakpointAction, role: BreakpointRole = BreakpointRole.CREATE
    ) -> None:
        self._builder._build_cmds_breakpoint(action, role)


class DebugConnection(BaseNetQASMConnection):
    """Connection that mocks most of the `BaseNetQASMConnection` logic.

    Subroutines that are flushed are simply stored in this object.
    No actual connection is made.
    """

    node_ids: Dict[str, int] = {}

    def __init__(self, *args, **kwargs):
        """A connection that simply stores the subroutine it commits"""
        self.storage = []
        super().__init__(*args, **kwargs)

    @property
    def shared_memory(self) -> SharedMemory:
        return SharedMemory()

    def _commit_serialized_message(
        self, raw_msg: bytes, block: bool = True, callback: Optional[Callable] = None
    ) -> None:
        """Commit a message to the backend/qnodeos"""
        self.storage.append(raw_msg)

    def _get_network_info(self) -> Type[NetworkInfo]:
        return DebugNetworkInfo


class DebugNetworkInfo(NetworkInfo):
    @classmethod
    def _get_node_id(cls, node_name: str) -> int:
        """Returns the node id for the node with the given name"""
        node_id = DebugConnection.node_ids.get(node_name)
        if node_id is None:
            raise ValueError(f"{node_name} is not a known node name")
        return node_id

    @classmethod
    def _get_node_name(cls, node_id: int) -> str:
        """Returns the node name for the node with the given ID"""
        for n_name, n_id in DebugConnection.node_ids.items():
            if n_id == node_id:
                return n_name
        raise ValueError(f"{node_id} is not a known node ID")

    @classmethod
    def get_node_id_for_app(cls, app_name: str) -> int:
        """Returns the node id for the app with the given name"""
        return cls._get_node_id(node_name=app_name)

    @classmethod
    def get_node_name_for_app(cls, app_name: str) -> str:
        """Returns the node name for the app with the given name"""
        return app_name
